import math
from unittest import TestCase
from hjson.compat import long_type, text_type
import hjson as json
from hjson.decoder import NaN, PosInf, NegInf

class TestFloat(TestCase):

    def test_degenerates_ignore(self):
        for f in (PosInf, NegInf, NaN):
            self.assertEqual(json.loads(json.dumpsJSON(f)), None)

    def test_floats(self):
        for num in [1617161771.7650001, math.pi, math.pi**100,
                    math.pi**-100, 3.1]:
            self.assertEqual(float(json.dumpsJSON(num)), num)
            self.assertEqual(json.loads(json.dumpsJSON(num)), num)
            self.assertEqual(json.loads(text_type(json.dumpsJSON(num))), num)

    def test_ints(self):
        for num in [1, long_type(1), 1<<32, 1<<64]:
            self.assertEqual(json.dumpsJSON(num), str(num))
            self.assertEqual(int(json.dumpsJSON(num)), num)
            self.assertEqual(json.loads(json.dumpsJSON(num)), num)
            self.assertEqual(json.loads(text_type(json.dumpsJSON(num))), num)



# © Copyright Databand.ai, an IBM Company 2022

from targets import DataTarget


class DataTransaction(object):
    def __init__(self, target):  # type: (DataTarget) -> DataTransaction
        super(DataTransaction, self).__init__()
        self.target = target
        # right now we are not having "atomic writes"
        # just marking the target as complete at the end

    def __enter__(self):
        return self.target

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            # There were no exceptions
            self.target.mark_success()
        return False  # False means we don't suppress the exception



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd._core.utils.traversing import flatten
from targets.data_target import DataTarget
from targets.target_config import TargetConfig
from targets.utils.open_multiple import MultiTargetOpen


logger = logging.getLogger(__name__)


class MultiTarget(DataTarget):
    def __init__(self, targets, properties=None, source=None):
        super(MultiTarget, self).__init__(properties=properties, source=source)
        self._targets = targets

        if targets and len({t.config for t in targets if hasattr(t, "config")}) == 1:
            self.config = targets[0].config
        else:
            self.config = TargetConfig()

    @property
    def targets(self):
        return self._targets

    def open(self, mode="r"):
        """ """
        if "r" in mode:
            return MultiTargetOpen(targets=self.list_partitions(), mode=mode)
        elif "w" in mode:
            raise NotImplementedError(
                "Multi part target can be used only for reading! You can not remove it!"
            )
        else:
            raise ValueError("Unsupported open mode '{}'".format(mode))

    def remove(self):
        raise NotImplementedError(
            "Multi part target can be used only for reading! You can not remove it!"
        )

    def exists(self):
        return all((t.exists() for t in flatten(self.targets) if t is not None))

    def list_partitions(self):
        return self.targets

    def __repr__(self):
        return "target('%s')" % ",".join(map(str, self.targets))

    def __str__(self):
        return ",".join(map(str, self.targets))



# © Copyright Databand.ai, an IBM Company 2022

import logging
import typing

import six

from dbnd._core.configuration.environ_config import is_databand_enabled
from dbnd._core.parameter.parameter_builder import parameter
from dbnd._core.task.decorated_callable_task import (
    DecoratedPipelineTask,
    DecoratedPythonTask,
)
from dbnd._core.task_build.task_decorator import (
    TaskDecorator,
    _UserClassWithTaskDecoratorMetaclass,
    build_dbnd_decorated_func,
)
from dbnd._core.task_build.task_registry import get_task_registry
from dbnd._core.tracking.managers.callable_tracking import _do_nothing_decorator


if typing.TYPE_CHECKING:
    pass

logger = logging.getLogger(__name__)
_default_output = parameter.output.pickle[object]


def task(*args, **kwargs):
    """
    Decorator that makes the following function a dbnd task.

    In DBND, a task is a modular unit of computation, usually a data transformation with an input and an output.
    To create a task, you need to decorate your function with the @task decorator.
    Under the hood, the task object is created, the function parameters are mapped to the task parameters,
    and the function returns value to task outputs.

    Example::

        @task
        def prepare_data(data: pd.DataFrame) -> pd.DataFrame:
            return data
    """
    kwargs.setdefault("_task_type", DecoratedPythonTask)
    kwargs.setdefault("_task_default_result", _default_output)
    return build_task_decorator(*args, **kwargs)


def pipeline(*args, **kwargs):
    """
    Decorator that makes the following function a dbnd pipeline.

    A pipeline is a collection of tasks wired together to run in a specific order.
    Pipelines define tasks and data flow.
    Pipelines encapsulate an execution plan that can run as a typical python function.

    Example::

        @pipeline
        def prepare_data_pipeline(data: pd.DataFrame):
            prepared_data = prepare_data(data)
            model = train_model(prepared_data)
            return model
    """
    kwargs.setdefault("_task_type", DecoratedPipelineTask)
    kwargs.setdefault("_task_default_result", parameter.output)
    return build_task_decorator(*args, **kwargs)


band = pipeline
"""
This is the method you should override while using PipelineTask.

Your Pipeline.band() call should have one or more tasks wired one into another.

Example::

    class PrepareData(PipelineTask):
        data = parameter.data
        prepared_data = output.csv.data

        def band(self):
            self.prepared_data = gather_data.dbnd_run()"""

data_source_pipeline = pipeline


def build_task_decorator(*decorator_args, **decorator_kwargs):
    # this code creates a new decorator that can be applied on any User Code

    if not is_databand_enabled():
        # simple `@task` decorator, no options were (probably) given.
        if len(decorator_args) == 1 and callable(decorator_args[0]):
            return decorator_args[0]
        return _do_nothing_decorator

    def class_or_func_decorator(class_or_func):
        # this code will run during compile time, when we apply dbnd decorator (for example: @task)
        task_decorator = TaskDecorator(class_or_func, decorator_kwargs=decorator_kwargs)
        tp = task_decorator.task_passport

        # we need to manually register the task here, since in regular flow
        # this happens in TaskMetaclass, but it's not invoked here due to lazy
        # evaluation task_cls
        r = get_task_registry()
        r.register_task_cls_factory(
            task_cls_factory=task_decorator.get_task_cls,
            full_task_family=tp.full_task_family,
            task_family=tp.task_family,
        )
        if task_decorator.is_class:
            # we will change metaclass for UserClass so we will process all UserClass calls
            #
            # @task
            # class UserClass():
            #     pass
            # so the the moment user call UserClass(), -> _DecoratedUserClassMeta.__call__ will be called
            dbnd_decorated_class = six.add_metaclass(
                _UserClassWithTaskDecoratorMetaclass
            )(class_or_func)
            dbnd_decorated_class.task_decorator = task_decorator
            task_decorator.class_or_func = dbnd_decorated_class
            return dbnd_decorated_class
        else:
            # @task
            # def user_func():
            #     pass
            # we will return our wrapper, that will be called during a runtime,
            # when user calls his own code.
            return build_dbnd_decorated_func(task_decorator)

    # simple `@task` decorator in opposite to @task(...), no options were (probably) given.
    if len(decorator_args) == 1 and callable(decorator_args[0]):
        return class_or_func_decorator(decorator_args[0])

    return class_or_func_decorator



# © Copyright Databand.ai, an IBM Company 2022

import hashlib
import re

import attr

from dbnd._core.utils import json_utils
from dbnd._core.utils.traversing import flatten, traverse_frozen_set


SIGNATURE_ID_TRUNCATE_HASH = 10

TASK_ID_INCLUDE_PARAMS = 3
TASK_ID_TRUNCATE_PARAMS = 16
TASK_ID_INVALID_CHAR_REGEX = re.compile(r"[^A-Za-z0-9_.]")


@attr.s
class Signature(object):
    name = attr.ib()  # type: str
    signature = attr.ib()  # type: str
    signature_source = attr.ib()  # type: str


def task_params_short(params):
    param_summary = "_".join(
        p[:TASK_ID_TRUNCATE_PARAMS]
        for p in (params[p] for p in sorted(params)[:TASK_ID_INCLUDE_PARAMS])
    )
    param_summary = TASK_ID_INVALID_CHAR_REGEX.sub("_", param_summary)

    return param_summary


def build_signature(name, params, extra=None):
    """
    Returns a canonical string used to identify a particular task

    :param name:
    :param params: a list mapping parameter names to their serialized values
    :return: A unique, shortened identifier corresponding to the family and params
    """
    # task_id is a concatenation of task family,
    # sorted by parameter name and a md5hash of the family/parameters as a cananocalised json.

    params = {key: value for key, value in params}

    signature_dict = {"name": name, "params": params}

    if extra:
        signature_dict["extra"] = extra

    # we can't handle sets
    signature_dict = traverse_frozen_set(signature_dict)
    param_str = json_utils.dumps_canonical(signature_dict)

    signature = user_friendly_signature(param_str)

    return Signature(name=name, signature=signature, signature_source=param_str)


def build_signature_from_values(name, struct):
    values = set([str(value) for value in flatten(struct)])
    signature_str = "|".join(sorted(values))
    signature = hashlib.md5(signature_str.encode("utf-8")).hexdigest()  # nosec B324
    return Signature(name=name, signature=signature, signature_source=signature_str)


def user_friendly_signature(str_value):
    hash_md5 = hashlib.md5(str_value.encode("utf-8")).hexdigest()  # nosec B324

    return hash_md5[:SIGNATURE_ID_TRUNCATE_HASH]



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.errors.base import DatabandConnectionException


def api_connection_refused(connection_details, ex):
    return DatabandConnectionException(
        "Connection Error: Could not connect to databand api server on host: %s "
        % connection_details,
        show_exc_info=False,
        nested_exceptions=[ex],
        help_msg="1. Please check that databand server is up and accessible from client side.\n"
        "2. You can temporary disable web tracking using --disable-web-tracker",
    )


def run_uid_not_found_in_db(run_uid):
    return DatabandConnectionException(
        "Could not find run uid '%s' in current database, are you using same DB for submission and execution?"
        % run_uid,
        show_exc_info=False,
        help_msg="Please check your configuration files, may be you are using local DB\n "
        "Validate that you don't have reference to previous run via --run-driver [UID] ",
    )


def couldnt_find_databand_run_in_db(name_or_uid, ex):
    return DatabandConnectionException(
        "Couldn't find the databand run '%s' in current database! Please validate that you are using the correct "
        "connection string" % name_or_uid,
        nested_exceptions=ex,
    )



# © Copyright Databand.ai, an IBM Company 2022

# inspired by wrapt.CallableObjectProxy
class CallableLazyObjectProxy(object):
    """
    Will proxy all attributes access and __call__ invocation to wrapped object
    returned by provided function.

    Provided function is evaluated only on access, allowing lazy evaluation.
    Provided function is evaluated each time to allow changing underlying
    wrapped object.
    """

    __slots__ = ["__wrapped__", "__weakref__"]

    def __init__(self, wrapped):
        self.__wrapped__ = wrapped

    @property
    def _wrapped(self):
        return self.__wrapped__()

    @property
    def __name__(self):
        return self._wrapped.__name__

    @property
    def __class__(self):
        return self._wrapped.__class__

    @property
    def __annotations__(self):
        return self._wrapped.__annotations__

    def __str__(self):
        return str(self._wrapped)

    def __repr__(self):
        return "<{} at 0x{:x} for {} at 0x{:x}>".format(
            type(self).__name__,
            id(self),
            type(self._wrapped).__name__,
            id(self._wrapped),
        )

    def __getattr__(self, item):
        return getattr(self._wrapped, item)

    def __call__(self, *args, **kwargs):
        return self._wrapped(*args, **kwargs)



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.errors import DatabandBuildError
from dbnd._core.errors.errors_utils import safe_value


def unknown_value_type_in_parameter(type_):
    from targets.values import get_types_registry

    return DatabandBuildError(
        "The parameter type '{}' is unknown, ".format(type_),
        help_msg="Use one of: {}, or register the new one via  'register_custom_parameter'".format(
            get_types_registry().list_known_types()
        ),
    )


def no_value_type_defined_in_parameter(context):
    from targets.values import get_types_registry

    return DatabandBuildError(
        "The parameter {context} doesn't have type! Please use parameter[YOUR_TYPE] or parameter[object]".format(
            context=context
        ),
        help_msg="Use one of: {}, or register the new type via  'register_custom_parameter'".format(
            get_types_registry().list_known_types()
        ),
    )


def no_value_type_from_default(default_value, context):
    from targets.values import get_types_registry

    return DatabandBuildError(
        "The parameter '{parameter}' has default value '{default_value}' "
        "that cannot be resolved into known type!"
        "Please use '{parameter} = parameter[YOUR_TYPE]' or '{parameter} = parameter[object]'".format(
            default_value=default_value, parameter=context
        ),
        help_msg="Use one of: {}, or register the new type via  'register_custom_parameter'".format(
            get_types_registry().list_known_types()
        ),
    )


def result_and_params_have_same_keys(context, conflict_keys):
    return DatabandBuildError(
        "{context} have same keys in result schema and function args: {conflict_keys}".format(
            context=context, conflict_keys=conflict_keys
        ),
        help_msg="Please, check %s and rename conflicted keys" % context,
    )


def task_env_param_with_no_env(context, key):
    return DatabandBuildError(
        "{context} can't calculate {key} and env is not defined".format(
            context=context, key=key
        ),
        help_msg="Please, check that %s has task_env parameter" % context,
    )


def task_env_param_not_exists_in_env(context, key, env_config):
    return DatabandBuildError(
        "{context} can't calculate {key}, as it doesn't exists in {env_config}".format(
            context=context, key=key, env_config=env_config
        ),
        help_msg="Please, check {context} definition, only parameters defined at {env_config} "
        "can have 'from_task_env_config=True'".format(
            context=context, env_config=env_config
        ),
    )


def failed_to_build_parameter(context, parameter_state, ex):
    return DatabandBuildError(
        "Failed to build parameter '{context}' defined as '{parameter}': {ex}".format(
            context=context, parameter=parameter_state, ex=ex
        ),
        help_msg="Validate parameter implementation at {}".format(context),
        nested_exceptions=[ex],
    )


def failed_to_convert_to_target_type(param, x, ex):
    return DatabandBuildError(
        "Failed to convert '{value}' to {param_type}".format(
            value=safe_value(x), param=param, param_type=param.value_type.type
        ),
        show_exc_info=False,
        nested_exceptions=[ex],
        help_msg="Check your %s logic" % param.name,
    )


def value_is_not_parameter_cls(value, context=None):
    return DatabandBuildError(
        "Parameter {context} is '{value}', while it should be defined via 'parameter'".format(
            value=value, context=context
        ),
        help_msg="Please, check your implementation (see relevant lines in exception message)",
    )


def sub_type_with_non_structural_value(context, value_type, sub_type):
    return DatabandBuildError(
        "Sub type {sub_type} is not supported by main value type :{value_type}.".format(
            sub_type=sub_type, value_type=value_type
        ),
        help_msg=" Use main value type like List/Set/Dict: {context} = parameter.sub_type(int)[List]",
    )


def dict_in_result_definition(result_deco):
    return DatabandBuildError(
        "Result definition should be tuple/list, we don't support dict in definition, got: {}".format(
            result_deco
        ),
        help_msg="You can use dict for output of the task "
        "{'features': some_output, 'scores': some_outputs} "
        "but not in result=<DEFINITION>. For example: result=('features', 'scores')",
    )


def wrong_result_definition(result_deco):
    return DatabandBuildError(
        "{} should be Parameter of task_output kind , "
        "or one of list/tuple of parameters or str".format(result_deco)
    )



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd._core.tracking.backends.channels.abstract_channel import TrackingChannel


logger = logging.getLogger(__name__)


class DisabledTrackingChannel(TrackingChannel):
    def __init__(self, *args, **kwargs):
        super(DisabledTrackingChannel, self).__init__(*args, **kwargs)
        logger.info("Tracking store is disable at core.tracker_api.")

    def is_ready(self):
        return True

    def _handle(self, name, data):
        pass



# index.py -- File parser/writer for the git index file
# Copyright (C) 2008-2013 Jelmer Vernooij <jelmer@jelmer.uk>
#
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as public by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Parser for the git index file format."""

import collections
import errno
import os
import stat
import struct
import sys

from dbnd._vendor.dulwich.file import GitFile
from dbnd._vendor.dulwich.objects import (
    Blob,
    S_IFGITLINK,
    S_ISGITLINK,
    Tree,
    hex_to_sha,
    sha_to_hex,
)
from dbnd._vendor.dulwich.pack import (
    SHA1Reader,
    SHA1Writer,
)


IndexEntry = collections.namedtuple(
    "IndexEntry",
    ["ctime", "mtime", "dev", "ino", "mode", "uid", "gid", "size", "sha", "flags"],
)


FLAG_STAGEMASK = 0x3000
FLAG_VALID = 0x8000
FLAG_EXTENDED = 0x4000


def pathsplit(path):
    """Split a /-delimited path into a directory part and a basename.

    :param path: The path to split.
    :return: Tuple with directory name and basename
    """
    try:
        (dirname, basename) = path.rsplit(b"/", 1)
    except ValueError:
        return (b"", path)
    else:
        return (dirname, basename)


def pathjoin(*args):
    """Join a /-delimited path.

    """
    return b"/".join([p for p in args if p])


def read_cache_time(f):
    """Read a cache time.

    :param f: File-like object to read from
    :return: Tuple with seconds and nanoseconds
    """
    return struct.unpack(">LL", f.read(8))


def write_cache_time(f, t):
    """Write a cache time.

    :param f: File-like object to write to
    :param t: Time to write (as int, float or tuple with secs and nsecs)
    """
    if isinstance(t, int):
        t = (t, 0)
    elif isinstance(t, float):
        (secs, nsecs) = divmod(t, 1.0)
        t = (int(secs), int(nsecs * 1000000000))
    elif not isinstance(t, tuple):
        raise TypeError(t)
    f.write(struct.pack(">LL", *t))


def read_cache_entry(f):
    """Read an entry from a cache file.

    :param f: File-like object to read from
    :return: tuple with: device, inode, mode, uid, gid, size, sha, flags
    """
    beginoffset = f.tell()
    ctime = read_cache_time(f)
    mtime = read_cache_time(f)
    (dev, ino, mode, uid, gid, size, sha, flags,) = struct.unpack(
        ">LLLLLL20sH", f.read(20 + 4 * 6 + 2)
    )
    name = f.read((flags & 0x0FFF))
    # Padding:
    real_size = (f.tell() - beginoffset + 8) & ~7
    f.read((beginoffset + real_size) - f.tell())
    return (
        name,
        ctime,
        mtime,
        dev,
        ino,
        mode,
        uid,
        gid,
        size,
        sha_to_hex(sha),
        flags & ~0x0FFF,
    )


def write_cache_entry(f, entry):
    """Write an index entry to a file.

    :param f: File object
    :param entry: Entry to write, tuple with:
        (name, ctime, mtime, dev, ino, mode, uid, gid, size, sha, flags)
    """
    beginoffset = f.tell()
    (name, ctime, mtime, dev, ino, mode, uid, gid, size, sha, flags) = entry
    write_cache_time(f, ctime)
    write_cache_time(f, mtime)
    flags = len(name) | (flags & ~0x0FFF)
    f.write(
        struct.pack(
            b">LLLLLL20sH",
            dev & 0xFFFFFFFF,
            ino & 0xFFFFFFFF,
            mode,
            uid,
            gid,
            size,
            hex_to_sha(sha),
            flags,
        )
    )
    f.write(name)
    real_size = (f.tell() - beginoffset + 8) & ~7
    f.write(b"\0" * ((beginoffset + real_size) - f.tell()))


def read_index(f):
    """Read an index file, yielding the individual entries."""
    header = f.read(4)
    if header != b"DIRC":
        raise AssertionError("Invalid index file header: %r" % header)
    (version, num_entries) = struct.unpack(b">LL", f.read(4 * 2))
    assert version in (1, 2)
    for i in range(num_entries):
        yield read_cache_entry(f)


def read_index_dict(f):
    """Read an index file and return it as a dictionary.

    :param f: File object to read from
    """
    ret = {}
    for x in read_index(f):
        ret[x[0]] = IndexEntry(*x[1:])
    return ret


def write_index(f, entries):
    """Write an index file.

    :param f: File-like object to write to
    :param entries: Iterable over the entries to write
    """
    f.write(b"DIRC")
    f.write(struct.pack(b">LL", 2, len(entries)))
    for x in entries:
        write_cache_entry(f, x)


def write_index_dict(f, entries):
    """Write an index file based on the contents of a dictionary.

    """
    entries_list = []
    for name in sorted(entries):
        entries_list.append((name,) + tuple(entries[name]))
    write_index(f, entries_list)


def cleanup_mode(mode):
    """Cleanup a mode value.

    This will return a mode that can be stored in a tree object.

    :param mode: Mode to clean up.
    """
    if stat.S_ISLNK(mode):
        return stat.S_IFLNK
    elif stat.S_ISDIR(mode):
        return stat.S_IFDIR
    elif S_ISGITLINK(mode):
        return S_IFGITLINK
    ret = stat.S_IFREG | 0o644
    ret |= mode & 0o111
    return ret


class Index(object):
    """A Git Index file."""

    def __init__(self, filename):
        """Open an index file.

        :param filename: Path to the index file
        """
        self._filename = filename
        self.clear()
        self.read()

    @property
    def path(self):
        return self._filename

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self._filename)

    def write(self):
        """Write current contents of index to disk."""
        f = GitFile(self._filename, "wb")
        try:
            f = SHA1Writer(f)
            write_index_dict(f, self._byname)
        finally:
            f.close()

    def read(self):
        """Read current contents of index from disk."""
        if not os.path.exists(self._filename):
            return
        f = GitFile(self._filename, "rb")
        try:
            f = SHA1Reader(f)
            for x in read_index(f):
                self[x[0]] = IndexEntry(*x[1:])
            # FIXME: Additional data?
            f.read(os.path.getsize(self._filename) - f.tell() - 20)
            f.check_sha()
        finally:
            f.close()

    def __len__(self):
        """Number of entries in this index file."""
        return len(self._byname)

    def __getitem__(self, name):
        """Retrieve entry by relative path.

        :return: tuple with (ctime, mtime, dev, ino, mode, uid, gid, size, sha,
            flags)
        """
        return self._byname[name]

    def __iter__(self):
        """Iterate over the paths in this index."""
        return iter(self._byname)

    def get_sha1(self, path):
        """Return the (git object) SHA1 for the object at a path."""
        return self[path].sha

    def get_mode(self, path):
        """Return the POSIX file mode for the object at a path."""
        return self[path].mode

    def iterobjects(self):
        """Iterate over path, sha, mode tuples for use with commit_tree."""
        for path in self:
            entry = self[path]
            yield path, entry.sha, cleanup_mode(entry.mode)

    def iterblobs(self):
        import warnings

        warnings.warn("Use iterobjects() instead.", PendingDeprecationWarning)
        return self.iterobjects()

    def clear(self):
        """Remove all contents from this index."""
        self._byname = {}

    def __setitem__(self, name, x):
        assert isinstance(name, bytes)
        assert len(x) == 10
        # Remove the old entry if any
        self._byname[name] = IndexEntry(*x)

    def __delitem__(self, name):
        assert isinstance(name, bytes)
        del self._byname[name]

    def iteritems(self):
        return self._byname.items()

    def items(self):
        return self._byname.items()

    def update(self, entries):
        for name, value in entries.items():
            self[name] = value

    def changes_from_tree(self, object_store, tree, want_unchanged=False):
        """Find the differences between the contents of this index and a tree.

        :param object_store: Object store to use for retrieving tree contents
        :param tree: SHA1 of the root tree
        :param want_unchanged: Whether unchanged files should be reported
        :return: Iterator over tuples with (oldpath, newpath), (oldmode,
            newmode), (oldsha, newsha)
        """

        def lookup_entry(path):
            entry = self[path]
            return entry.sha, entry.mode

        for (name, mode, sha) in changes_from_tree(
            self._byname.keys(),
            lookup_entry,
            object_store,
            tree,
            want_unchanged=want_unchanged,
        ):
            yield (name, mode, sha)

    def commit(self, object_store):
        """Create a new tree from an index.

        :param object_store: Object store to save the tree in
        :return: Root tree SHA
        """
        return commit_tree(object_store, self.iterobjects())


def commit_tree(object_store, blobs):
    """Commit a new tree.

    :param object_store: Object store to add trees to
    :param blobs: Iterable over blob path, sha, mode entries
    :return: SHA1 of the created tree.
    """

    trees = {b"": {}}

    def add_tree(path):
        if path in trees:
            return trees[path]
        dirname, basename = pathsplit(path)
        t = add_tree(dirname)
        assert isinstance(basename, bytes)
        newtree = {}
        t[basename] = newtree
        trees[path] = newtree
        return newtree

    for path, sha, mode in blobs:
        tree_path, basename = pathsplit(path)
        tree = add_tree(tree_path)
        tree[basename] = (mode, sha)

    def build_tree(path):
        tree = Tree()
        for basename, entry in trees[path].items():
            if isinstance(entry, dict):
                mode = stat.S_IFDIR
                sha = build_tree(pathjoin(path, basename))
            else:
                (mode, sha) = entry
            tree.add(basename, mode, sha)
        object_store.add_object(tree)
        return tree.id

    return build_tree(b"")


def commit_index(object_store, index):
    """Create a new tree from an index.

    :param object_store: Object store to save the tree in
    :param index: Index file
    :note: This function is deprecated, use index.commit() instead.
    :return: Root tree sha.
    """
    return commit_tree(object_store, index.iterobjects())


def changes_from_tree(names, lookup_entry, object_store, tree, want_unchanged=False):
    """Find the differences between the contents of a tree and
    a working copy.

    :param names: Iterable of names in the working copy
    :param lookup_entry: Function to lookup an entry in the working copy
    :param object_store: Object store to use for retrieving tree contents
    :param tree: SHA1 of the root tree, or None for an empty tree
    :param want_unchanged: Whether unchanged files should be reported
    :return: Iterator over tuples with (oldpath, newpath), (oldmode, newmode),
        (oldsha, newsha)
    """
    # TODO(jelmer): Support a include_trees option
    other_names = set(names)

    if tree is not None:
        for (name, mode, sha) in object_store.iter_tree_contents(tree):
            try:
                (other_sha, other_mode) = lookup_entry(name)
            except KeyError:
                # Was removed
                yield ((name, None), (mode, None), (sha, None))
            else:
                other_names.remove(name)
                if want_unchanged or other_sha != sha or other_mode != mode:
                    yield ((name, name), (mode, other_mode), (sha, other_sha))

    # Mention added files
    for name in other_names:
        try:
            (other_sha, other_mode) = lookup_entry(name)
        except KeyError:
            pass
        else:
            yield ((None, name), (None, other_mode), (None, other_sha))


def index_entry_from_stat(stat_val, hex_sha, flags, mode=None):
    """Create a new index entry from a stat value.

    :param stat_val: POSIX stat_result instance
    :param hex_sha: Hex sha of the object
    :param flags: Index flags
    """
    if mode is None:
        mode = cleanup_mode(stat_val.st_mode)

    return IndexEntry(
        stat_val.st_ctime,
        stat_val.st_mtime,
        stat_val.st_dev,
        stat_val.st_ino,
        mode,
        stat_val.st_uid,
        stat_val.st_gid,
        stat_val.st_size,
        hex_sha,
        flags,
    )


def build_file_from_blob(blob, mode, target_path, honor_filemode=True):
    """Build a file or symlink on disk based on a Git object.

    :param obj: The git object
    :param mode: File mode
    :param target_path: Path to write to
    :param honor_filemode: An optional flag to honor core.filemode setting in
        config file, default is core.filemode=True, change executable bit
    :return: stat object for the file
    """
    try:
        oldstat = os.lstat(target_path)
    except OSError as e:
        if e.errno == errno.ENOENT:
            oldstat = None
        else:
            raise
    contents = blob.as_raw_string()
    if stat.S_ISLNK(mode):
        # FIXME: This will fail on Windows. What should we do instead?
        if oldstat:
            os.unlink(target_path)
        if sys.platform == "win32" and sys.version_info[0] == 3:
            # os.readlink on Python3 on Windows requires a unicode string.
            # TODO(jelmer): Don't assume tree_encoding == fs_encoding
            tree_encoding = sys.getfilesystemencoding()
            contents = contents.decode(tree_encoding)
            target_path = target_path.decode(tree_encoding)
        os.symlink(contents, target_path)
    else:
        if oldstat is not None and oldstat.st_size == len(contents):
            with open(target_path, "rb") as f:
                if f.read() == contents:
                    return oldstat

        with open(target_path, "wb") as f:
            # Write out file
            f.write(contents)

        if honor_filemode:
            os.chmod(target_path, mode)

    return os.lstat(target_path)


INVALID_DOTNAMES = (b".git", b".", b"..", b"")


def validate_path_element_default(element):
    return element.lower() not in INVALID_DOTNAMES


def validate_path_element_ntfs(element):
    stripped = element.rstrip(b". ").lower()
    if stripped in INVALID_DOTNAMES:
        return False
    if stripped == b"git~1":
        return False
    return True


def validate_path(path, element_validator=validate_path_element_default):
    """Default path validator that just checks for .git/."""
    parts = path.split(b"/")
    for p in parts:
        if not element_validator(p):
            return False
    else:
        return True


def build_index_from_tree(
    root_path,
    index_path,
    object_store,
    tree_id,
    honor_filemode=True,
    validate_path_element=validate_path_element_default,
):
    """Generate and materialize index from a tree

    :param tree_id: Tree to materialize
    :param root_path: Target dir for materialized index files
    :param index_path: Target path for generated index
    :param object_store: Non-empty object store holding tree contents
    :param honor_filemode: An optional flag to honor core.filemode setting in
        config file, default is core.filemode=True, change executable bit
    :param validate_path_element: Function to validate path elements to check
        out; default just refuses .git and .. directories.

    :note:: existing index is wiped and contents are not merged
        in a working dir. Suitable only for fresh clones.
    """

    index = Index(index_path)
    if not isinstance(root_path, bytes):
        root_path = root_path.encode(sys.getfilesystemencoding())

    for entry in object_store.iter_tree_contents(tree_id):
        if not validate_path(entry.path, validate_path_element):
            continue
        full_path = _tree_to_fs_path(root_path, entry.path)

        if not os.path.exists(os.path.dirname(full_path)):
            os.makedirs(os.path.dirname(full_path))

        # TODO(jelmer): Merge new index into working tree
        if S_ISGITLINK(entry.mode):
            if not os.path.isdir(full_path):
                os.mkdir(full_path)
            st = os.lstat(full_path)
            # TODO(jelmer): record and return submodule paths
        else:
            obj = object_store[entry.sha]
            st = build_file_from_blob(
                obj, entry.mode, full_path, honor_filemode=honor_filemode
            )
        # Add file to index
        if not honor_filemode or S_ISGITLINK(entry.mode):
            # we can not use tuple slicing to build a new tuple,
            # because on windows that will convert the times to
            # longs, which causes errors further along
            st_tuple = (
                entry.mode,
                st.st_ino,
                st.st_dev,
                st.st_nlink,
                st.st_uid,
                st.st_gid,
                st.st_size,
                st.st_atime,
                st.st_mtime,
                st.st_ctime,
            )
            st = st.__class__(st_tuple)
        index[entry.path] = index_entry_from_stat(st, entry.sha, 0)

    index.write()


def blob_from_path_and_stat(fs_path, st):
    """Create a blob from a path and a stat object.

    :param fs_path: Full file system path to file
    :param st: A stat object
    :return: A `Blob` object
    """
    assert isinstance(fs_path, bytes)
    blob = Blob()
    if not stat.S_ISLNK(st.st_mode):
        with open(fs_path, "rb") as f:
            blob.data = f.read()
    else:
        if sys.platform == "win32" and sys.version_info[0] == 3:
            # os.readlink on Python3 on Windows requires a unicode string.
            # TODO(jelmer): Don't assume tree_encoding == fs_encoding
            tree_encoding = sys.getfilesystemencoding()
            fs_path = fs_path.decode(tree_encoding)
            blob.data = os.readlink(fs_path).encode(tree_encoding)
        else:
            blob.data = os.readlink(fs_path)
    return blob


def read_submodule_head(path):
    """Read the head commit of a submodule.

    :param path: path to the submodule
    :return: HEAD sha, None if not a valid head/repository
    """
    from dbnd._vendor.dulwich.errors import NotGitRepository
    from dbnd._vendor.dulwich.repo import Repo

    # Repo currently expects a "str", so decode if necessary.
    # TODO(jelmer): Perhaps move this into Repo() ?
    if not isinstance(path, str):
        path = path.decode(sys.getfilesystemencoding())
    try:
        repo = Repo(path)
    except NotGitRepository:
        return None
    try:
        return repo.head()
    except KeyError:
        return None


def get_unstaged_changes(index, root_path):
    """Walk through an index and check for differences against working tree.

    :param index: index to check
    :param root_path: path in which to find files
    :return: iterator over paths with unstaged changes
    """
    # For each entry in the index check the sha1 & ensure not staged
    if not isinstance(root_path, bytes):
        root_path = root_path.encode(sys.getfilesystemencoding())

    for tree_path, entry in index.iteritems():
        full_path = _tree_to_fs_path(root_path, tree_path)
        try:
            blob = blob_from_path_and_stat(full_path, os.lstat(full_path))
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
            # The file was removed, so we assume that counts as
            # different from whatever file used to exist.
            yield tree_path
        except IOError as e:
            if e.errno != errno.EISDIR:
                raise
            # This is actually a directory
            if os.path.exists(os.path.join(tree_path, ".git")):
                # Submodule
                head = read_submodule_head(tree_path)
                if entry.sha != head:
                    yield tree_path
            else:
                # The file was changed to a directory, so consider it removed.
                yield tree_path
        else:
            if blob.id != entry.sha:
                yield tree_path


os_sep_bytes = os.sep.encode("ascii")


def _tree_to_fs_path(root_path, tree_path):
    """Convert a git tree path to a file system path.

    :param root_path: Root filesystem path
    :param tree_path: Git tree path as bytes

    :return: File system path.
    """
    assert isinstance(tree_path, bytes)
    if os_sep_bytes != b"/":
        sep_corrected_path = tree_path.replace(b"/", os_sep_bytes)
    else:
        sep_corrected_path = tree_path
    return os.path.join(root_path, sep_corrected_path)


def _fs_to_tree_path(fs_path, fs_encoding=None):
    """Convert a file system path to a git tree path.

    :param fs_path: File system path.
    :param fs_encoding: File system encoding

    :return:  Git tree path as bytes
    """
    if fs_encoding is None:
        fs_encoding = sys.getfilesystemencoding()
    if not isinstance(fs_path, bytes):
        fs_path_bytes = fs_path.encode(fs_encoding)
    else:
        fs_path_bytes = fs_path
    if os_sep_bytes != b"/":
        tree_path = fs_path_bytes.replace(os_sep_bytes, b"/")
    else:
        tree_path = fs_path_bytes
    return tree_path


def index_entry_from_path(path, object_store=None):
    """Create an index from a filesystem path.

    This returns an index value for files, symlinks
    and tree references. for directories and
    non-existant files it returns None

    :param path: Path to create an index entry for
    :param object_store: Optional object store to
        save new blobs in
    :return: An index entry
    """
    assert isinstance(path, bytes)
    try:
        st = os.lstat(path)
        blob = blob_from_path_and_stat(path, st)
    except EnvironmentError as e:
        if e.errno == errno.EISDIR:
            if os.path.exists(os.path.join(path, b".git")):
                head = read_submodule_head(path)
                if head is None:
                    return None
                return index_entry_from_stat(st, head, 0, mode=S_IFGITLINK)
            else:
                raise
        else:
            raise
    else:
        if object_store is not None:
            object_store.add_object(blob)
        return index_entry_from_stat(st, blob.id, 0)


def iter_fresh_entries(paths, root_path, object_store=None):
    """Iterate over current versions of index entries on disk.

    :param paths: Paths to iterate over
    :param root_path: Root path to access from
    :param store: Optional store to save new blobs in
    :return: Iterator over path, index_entry
    """
    for path in paths:
        p = _tree_to_fs_path(root_path, path)
        try:
            entry = index_entry_from_path(p, object_store=object_store)
        except EnvironmentError as e:
            if e.errno in (errno.ENOENT, errno.EISDIR):
                entry = None
            else:
                raise
        yield path, entry


def iter_fresh_blobs(index, root_path):
    """Iterate over versions of blobs on disk referenced by index.

    Don't use this function; it removes missing entries from index.

    :param index: Index file
    :param root_path: Root path to access from
    :param include_deleted: Include deleted entries with sha and
        mode set to None
    :return: Iterator over path, sha, mode
    """
    import warnings

    warnings.warn(PendingDeprecationWarning, "Use iter_fresh_objects instead.")
    for entry in iter_fresh_objects(index, root_path, include_deleted=True):
        if entry[1] is None:
            del index[entry[0]]
        else:
            yield entry


def iter_fresh_objects(paths, root_path, include_deleted=False, object_store=None):
    """Iterate over versions of objecs on disk referenced by index.

    :param index: Index file
    :param root_path: Root path to access from
    :param include_deleted: Include deleted entries with sha and
        mode set to None
    :param object_store: Optional object store to report new items to
    :return: Iterator over path, sha, mode
    """
    for path, entry in iter_fresh_entries(paths, root_path, object_store=object_store):
        if entry is None:
            if include_deleted:
                yield path, None, None
        else:
            entry = IndexEntry(*entry)
            yield path, entry.sha, cleanup_mode(entry.mode)


def refresh_index(index, root_path):
    """Refresh the contents of an index.

    This is the equivalent to running 'git commit -a'.

    :param index: Index to update
    :param root_path: Root filesystem path
    """
    for path, entry in iter_fresh_entries(index, root_path):
        index[path] = path



# © Copyright Databand.ai, an IBM Company 2022

import abc
import copy
import logging
import re
import typing

import six

from dbnd._core.errors import friendly_error
from dbnd._core.utils.basics.load_python_module import run_user_func
from dbnd._vendor import fast_hasher
from targets.value_meta import ValueMeta


if typing.TYPE_CHECKING:
    from typing import Any, Union

    from targets import DataTarget, Target
    from targets.value_meta import ValueMetaConf

    T = typing.TypeVar("T")


logger = logging.getLogger(__name__)
_IS_INSTANCE_SAMPLE = 1000

_GLOB_PATH_REGEX = re.compile(r".*{.*,.*}.*")


def _is_glob_path(path):
    match = _GLOB_PATH_REGEX.match(path)
    return match is not None


class ValueType(object):
    support_merge = False
    support_from_str = True

    type_str_extras = None

    load_on_build = True
    discoverable = True  # we can discover it from the object

    is_lazy_evaluated = False

    @property
    @abc.abstractmethod
    def type(self):
        pass

    @property
    def type_str(self):
        if hasattr(self.type, "__name__"):
            s = str(self.type.__name__)
        else:
            s = str(self.type)

        return re.sub(r"^typing\.", "", s)

    @property
    def config_name(self):
        return "object"
        # if self.type is None:
        #     raise Exception("Value type doesn't have a name! %s" % self)
        # if hasattr(self.type, "__name__"):
        #     return str(self.type.__name__)
        # return str(self.type)

    def merge_values(self, *values, **kwargs):
        pass

    def is_type_of(self, value):
        return self.type is not None and type(value) == self.type

    def is_handler_of_type(self, type_):
        return self.type is not None and self.type == type_

    #################
    # transformations
    #################
    def normalize(self, x):  # type: (T) -> T
        """
        Given a parsed parameter value, normalizes it.

        The value can either be the result of parse(), the default value or
        arguments passed into the task's constructor by instantiation.

        This is very implementation defined, but can be used to validate/clamp
        valid values. For example, if you wanted to only accept even integers,
        and "correct" odd values to the nearest integer, you can implement
        normalize as ``x // 2 * 2``.
        """
        return x  # default impl

    def parse_from_str(self, x):  # type: (str) -> T
        if not self.support_from_str:
            raise friendly_error.targets.type_without_parse_from_str(self)
        return x  # default impl

    def parse_from_str_lines(self, lines):
        return self.parse_from_str("".join(lines))

    def parse_if_str(self, x):  # type: (Union[str, T]) -> T
        if isinstance(x, six.string_types):
            return self.parse_from_str(x)
        return x

    def _generate_empty_default(self):
        return None

    # dump

    def to_str(self, x):  # type: (T) -> str
        """
        Opposite of :py:meth:`parse`.

        Converts the value ``x`` to a string.

        :param x: the value to serialize.
        """
        return str(x)

    def to_str_lines(self, x):
        return self.to_str(x)

    def to_repr(self, x):
        return repr(x)

    def to_signature(self, x):
        return self.to_str(x)

    def to_preview(self, x, preview_size):  # type: (T, int) -> str
        return self.to_str(x)[:preview_size]

    ##################
    # Target I/O
    ##################

    def _interpolate_from_str(self, value):

        # when we parse string, if it's has @ - we always will want to "de-reference" it
        # and read from disk
        # it's very important for primitive types, as we don't know
        # if we need to parse("str") or read_from_disk("str")
        # @ in the beginning of the string supported by @@
        if isinstance(value, six.string_types) and value.startswith("@"):
            value = value[1:]  # we shift first @
            if value.startswith("@"):
                # this is the way to provide string value with @
                # we return value minus "@"
                return value
            # so this is @/somepath case and not @@
            if value.lower() == "none":
                return None

            from targets import target

            if value.startswith("python://"):
                value = value[9:]
                value = run_user_func(value)
            else:
                if value.startswith("target:"):
                    value = value[7:]
                    value = target(value)
                else:
                    # the path is a target, something like @/path/to/file/with/data
                    # this might be dangerous, we need to have more "explicit" way
                    # to understand that the path is a path
                    value = self.load_from_target(target(value))
        return value

    def parse_value(self, value, load_value=None, target_config=None, sub_value=False):
        """
        Parse an individual value from the input.

        probably this is the most important code in user value parsing
        :param str value: the value to parse.
        :return: the parsed value.
        """

        from dbnd._core.utils.task_utils import to_targets
        from targets import Target
        from targets.inmemory_target import InMemoryTarget
        from targets.values.target_values import _TargetValueType

        if load_value is None:
            load_value = self.load_on_build

        value = self._interpolate_from_str(value)
        if value is None:
            return value

        if isinstance(value, six.string_types):
            # we are in the string mode
            # it's can be "serialized to string" or path value
            if load_value:
                # in case we have simple type -> just load/parse it
                if self.support_from_str:
                    value = self.parse_from_str(value)
                    value = self.normalize(value)
                    return value

            # otherwise - the data is "Complex object"
            # our assumption is that it can not be loaded from string
            # the value is a path!
            target_kwargs = {}
            if target_config:
                target_kwargs["config"] = target_config

            # Check for glob path
            if _is_glob_path(value):
                from targets import target

                return target(value, config=target_config)

            """
            it's possible that we have a list of targets, or just a single target (all targets should be loaded as
            single object). we need to support:
                1. /some/path
                2. /some/path,....
                3. ["/some_path",..]
            we will try to parse it as list, if we get list with one element (1) -> we can  return it, otherwise we
            wrap it with MultiTarget
            """
            from targets.values.structure import ListValueType

            # Parse into value type list
            list_of_targets = ListValueType().parse_from_str(value)
            # Apply all values from config
            list_of_targets = to_targets(
                list_of_targets, from_string_kwargs=target_kwargs
            )

            if len(list_of_targets) == 1:
                return list_of_targets[0]
            else:
                from targets.multi_target import MultiTarget

                return MultiTarget(list_of_targets)

        from dbnd._core.task import Task

        if isinstance(value, Task):
            return to_targets(value)

        if isinstance(value, Target):
            return value

        # so we have a value that is obviously "Data" type,
        # we want to be able to support "load_value" behaviour
        if not load_value and not isinstance(self, _TargetValueType):
            return InMemoryTarget(value, value_type=self)

        value = self.normalize(value)
        return value

    def load_runtime(self, value):
        if not value:
            return value

        from targets import DataTarget, InMemoryTarget

        if isinstance(value, InMemoryTarget):
            return value.load()
        if isinstance(value, DataTarget):
            return self.load_from_target(value)
        return value

    def load_from_target(self, target, **kwargs):
        # type: (DataTarget, **Any) -> Any
        return target.load(value_type=self, **kwargs)

    def save_to_target(
        self, target, value, **kwargs
    ):  # type: (DataTarget, T, **Any)-> None
        target.dump(value, value_type=self, **kwargs)

    #####################
    # Utilities
    #####################
    def next_in_enumeration(self, _value):
        """
        If your Parameter type has an enumerable ordering of values. You can
        choose to override this method. This method is used by the
        :py:mod:`databand.execution_summary` module for pretty printing
        purposes. Enabling it to pretty print tasks like ``MyTask(num=1),
        MyTask(num=2), MyTask(num=3)`` to ``MyTask(num=1..3)``.

        :param _value: The value
        :return: The next value, like "value + 1". Or ``None`` if there's no enumerable ordering.
        """
        return None

    def __call__(self):
        """
        support putting instance of ValueType into Generic
        for example List[ValueType()]
        :return:
        """
        return self

    def __repr__(self):
        return "{self.type_str}@{self.__class__.__name__}".format(self=self)

    def __str__(self):
        return "{self.type_str}".format(self=self)

    def with_lazy_load(self, lazy_load=True):
        # a little bit confusing
        # load on build == not lazy_load
        if self.load_on_build != lazy_load:
            return self

        # copy doesn't work - bug in python generics
        new_self = copy.deepcopy(self)
        new_self.load_on_build = not lazy_load
        return new_self

    def with_sub_type_handler(self, type_handler):
        return self

    def get_value_meta(self, value, meta_conf):
        # type: (Any,  ValueMetaConf) -> ValueMeta

        if meta_conf.log_preview:
            preview = self.to_preview(value, preview_size=meta_conf.get_preview_size())
            data_hash = _safe_hash(value)
        else:
            preview = None
            data_hash = None

        data_schema = {"type": self.type_str}

        return ValueMeta(
            value_preview=preview,
            data_dimensions=None,
            data_schema=data_schema,
            data_hash=data_hash,
        )

    def support_fast_count(self, target):
        # type: (Target) -> bool
        return True


def _safe_hash(value):
    try:
        return fast_hasher.hash(value)
    except:
        logger.info("Failed to hash value of type %s", type(value))
        return None


class InlineValueType(ValueType):
    support_from_str = False

    def __init__(self, type_):
        self._type = type_

    @property
    def type(self):
        return self._type

    @property
    def type_str(self):
        try:
            return self._type.__name__
        except AttributeError:
            return self._type.__class__.__name__


def _isinstances(list_obj, type_):
    if not list_obj:
        return True
    if _IS_INSTANCE_SAMPLE and len(list_obj) > _IS_INSTANCE_SAMPLE:
        # we don't want to run isinstance on 1M objects..
        count = 0
        for elem in list_obj:
            if not isinstance(elem, type_):
                return False
            if count > _IS_INSTANCE_SAMPLE:
                return True
            count += 1
        return True

    return all(isinstance(elem, type_) for elem in list_obj)



# © Copyright Databand.ai, an IBM Company 2022

"""
Context: airflow operator/task is running a function with @task
Here we create dbnd objects to represent them and send to webserver through tracking api.
"""
import logging
import os
import uuid

from collections.abc import Mapping
from typing import Any, Dict, Iterable, Optional, Tuple
from uuid import UUID

import dbnd._core.utils.basics.environ_utils

from dbnd._core.configuration import environ_config
from dbnd._core.configuration.environ_config import (
    DBND_ROOT_RUN_UID,
    ENV_DBND_TRACKING_ATTEMPT_UID,
)
from dbnd._core.constants import UpdateSource
from dbnd._core.current import get_settings
from dbnd._core.task.tracking_task import TrackingTask
from dbnd._core.task_build.task_passport import TaskPassport
from dbnd._core.task_build.task_source_code import NO_SOURCE_CODE, TaskSourceCode
from dbnd._core.tracking.airflow_task_context import AirflowTaskContext
from dbnd._core.tracking.dbnd_spark_init import try_get_airflow_context_from_spark_conf
from dbnd._core.utils.airflow_cmd_utils import generate_airflow_cmd
from dbnd._core.utils.type_check_utils import is_instance_by_class_name
from dbnd._core.utils.uid_utils import (
    get_airflow_instance_uid,
    get_job_run_uid,
    get_task_def_uid,
    get_task_run_attempt_uid,
    get_task_run_uid,
    source_md5,
)


logger = logging.getLogger(__name__)


DAG_SPECIAL_TASK_ID = "DAG"


def override_airflow_log_system_for_tracking():
    return dbnd._core.utils.basics.environ_utils.environ_enabled(
        environ_config.ENV_DBND__OVERRIDE_AIRFLOW_LOG_SYSTEM_FOR_TRACKING
    )


def extract_airflow_context(airflow_context):
    # type: (Dict[str, Any]) -> Optional[AirflowTaskContext]
    """Create AirflowTaskContext for airflow_context dict"""

    task_instance = airflow_context.get("task_instance")
    if task_instance is None:
        return None

    dag_id = task_instance.dag_id
    task_id = task_instance.task_id
    execution_date = str(task_instance.execution_date)
    try_number = task_instance.try_number

    if dag_id and task_id and execution_date:
        return AirflowTaskContext(
            dag_id=dag_id,
            execution_date=execution_date,
            task_id=task_id,
            try_number=try_number,
            context=airflow_context,
            airflow_instance_uid=get_airflow_instance_uid(),
            is_subdag=airflow_context.get("dag").is_subdag,
        )

    logger.debug(
        "airflow context from inspect, at least one of those params is missing"
        "dag_id: {}, execution_date: {}, task_id: {}".format(
            dag_id, execution_date, task_id
        )
    )
    return None


def try_get_airflow_context():
    # type: ()-> Optional[AirflowTaskContext]
    # first try to get from spark, then from call stack and then from airflow env
    try:
        for func in [
            try_get_airflow_context_from_spark_conf,
            try_get_airflow_context_env,
        ]:
            context = func()
            if context:
                return context
            else:
                msg = func.__name__.replace("_", " ").replace("try", "couldn't")
                logger.debug(msg)

    except Exception:
        return None


def try_get_airflow_context_env():
    # type: ()-> Optional[AirflowTaskContext]
    # Those env vars are set by airflow before running the operator
    dag_id = os.environ.get("AIRFLOW_CTX_DAG_ID")
    execution_date = os.environ.get("AIRFLOW_CTX_EXECUTION_DATE")
    task_id = os.environ.get("AIRFLOW_CTX_TASK_ID")
    try_number = os.environ.get("AIRFLOW_CTX_TRY_NUMBER")
    airflow_instance_uid = os.environ.get("AIRFLOW_CTX_UID")

    if dag_id and task_id and execution_date:
        return AirflowTaskContext(
            dag_id=dag_id,
            execution_date=execution_date,
            task_id=task_id,
            try_number=int(try_number) if try_number else None,
            airflow_instance_uid=airflow_instance_uid,
        )

    logger.debug(
        "airflow context from env, at least one of those environment var is missing"
        "dag_id: {}, execution_date: {}, task_id: {}".format(
            dag_id, execution_date, task_id
        )
    )
    return None


def build_run_time_airflow_task(af_context, root_task_name):
    # type: (AirflowTaskContext, Optional[str]) -> Tuple[TrackingTask, str, UpdateSource, UUID]
    if af_context.context:
        # we are in the execute entry point and therefore that task name is <task>__execute
        task_family = af_context.task_id

        airflow_operator = af_context.context["task_instance"].task

        # find the template fields of the operators
        user_params = get_flatten_operator_params(airflow_operator)

        source_code = NO_SOURCE_CODE
        if is_instance_by_class_name(airflow_operator, "PythonOperator"):
            tracked_function = airflow_operator.python_callable
            user_params["function_name"] = tracked_function.__name__
            source_code = TaskSourceCode.from_callable(tracked_function)
    else:
        # if this is an inline run-time task, we name it after the script which ran it
        # If we ever get here, the root_task_name will be just "airflow" since this is what we pass
        task_family = get_task_family_for_inline_script(
            af_context.task_id, root_task_name
        )
        source_code = TaskSourceCode.from_callstack()
        user_params = {}

    user_params.update(
        dag_id=af_context.dag_id,
        execution_date=af_context.execution_date,
        task_version="%s:%s" % (af_context.task_id, af_context.execution_date),
    )

    # just a placeholder name
    task_passport = TaskPassport.from_module(task_family)
    task_definition_uid = get_task_def_uid(
        af_context.dag_id,
        task_family,
        "{}{}".format(
            source_md5(source_code.task_source_code),
            source_md5(source_code.task_module_code),
        ),
    )
    root_task = TrackingTask.for_user_params(
        task_definition_uid=task_definition_uid,
        task_name=task_family,
        task_passport=task_passport,
        source_code=source_code,
        user_params=user_params,
    )  # type: TrackingTask

    root_task.ctrl.task_repr.task_functional_call = ""
    root_task.ctrl.task_repr.task_command_line = generate_airflow_cmd(
        dag_id=af_context.dag_id,
        task_id=af_context.task_id,
        execution_date=af_context.execution_date,
        is_root_task=False,
    )

    root_run_uid = get_job_run_uid(
        airflow_instance_uid=af_context.airflow_instance_uid,
        dag_id=af_context.dag_id,
        execution_date=af_context.execution_date,
    )
    root_task.ctrl.force_task_run_uid = get_task_run_uid(
        run_uid=root_run_uid, dag_id=af_context.dag_id, task_id=task_family
    )

    job_name = af_context.dag_id
    source = UpdateSource.airflow_tracking
    return root_task, job_name, source, root_run_uid


def should_flatten(operator, attr_name):
    flatten_config = get_settings().tracking.flatten_operator_fields
    for op_name in flatten_config:
        if is_instance_by_class_name(operator, op_name):
            return attr_name in flatten_config[op_name]
    return False


def flatten_param(attr_name, value):
    # type: (str, Any) -> Iterable[Tuple[str, Any]]
    """
    Flatten means we take out the first layer of nested value and split it to elements with relevant name.
    examples:
    >>> flatten_param("name", [0,1, [1,2,3,4]]) == [("name.0", 0),  ("name.1", 1), ("name.2", [1,2,3,4]),]
    >>> flatten_param("other", {"a": 1, "b": {"inner": "value"}}) == [("other.a", 1),  ("other.b", {"inner": "value"})]
    >>> flatten_param("other_name", "1") == [("other_name", "1")]
    """
    if isinstance(value, (dict, Mapping)):
        for name, value in value.items():
            yield ("%s.%s" % (attr_name, name), value)

    elif isinstance(value, list):
        for i, value in enumerate(value):
            yield ("%s.%s" % (attr_name, i), value)

    else:
        yield (attr_name, value)


def get_flatten_operator_params(operator):
    params = []
    for attr_name in operator.template_fields:
        value = getattr(operator, attr_name)
        if should_flatten(operator, attr_name):
            params.extend(flatten_param(attr_name, value))

        else:
            params.append((attr_name, value))

    # we want to make sure that the value is str
    return {name: repr(value) for name, value in params}


def calc_task_key_from_af_ti(ti):
    """
    Creates a key from airflow TaskInstance, for communicating task_run_attempt_uid
    """
    return ":".join([ENV_DBND_TRACKING_ATTEMPT_UID, ti.dag_id, ti.task_id])


def get_task_family_for_inline_script(task_id, root_task_name):
    """
    Calculate task family name for inline script
    :param task_id: task_id
    :param root_task_name: root_task_name
    :return:
    """
    return "_".join([task_id, root_task_name])


def get_task_run_uid_for_inline_script(tracking_env, script_name):
    # type: (Dict[str, str], str) -> (str, UUID, UUID)
    """
    Calculate task run uid and task run attempt uid for inline script execution
    :param tracking_env: dict with airflow context env
    :param script_name: inline script name
    :return:
    """
    task_id = get_task_family_for_inline_script(
        tracking_env["AIRFLOW_CTX_TASK_ID"], script_name
    )
    run_uid = uuid.UUID(tracking_env[DBND_ROOT_RUN_UID])
    dag_id = tracking_env["AIRFLOW_CTX_DAG_ID"]
    task_run_uid = get_task_run_uid(run_uid, dag_id, task_id)
    task_run_attempt_uid = get_task_run_attempt_uid(
        run_uid, dag_id, task_id, tracking_env["AIRFLOW_CTX_TRY_NUMBER"]
    )
    return task_id, task_run_uid, task_run_attempt_uid



# © Copyright Databand.ai, an IBM Company 2022

import os
import pprint
import sys

from dbnd._core.tracking.dbnd_spark_init import (
    _is_dbnd_spark_installed,
    try_get_airflow_context_from_spark_conf,
    verify_spark_pre_conditions,
)


pp = pprint.PrettyPrinter(indent=4)


def custom_print(msg):
    print(msg, file=sys.stderr)


def _is_dbnd_airflow_key(k):
    return "dbnd" in k.lower() or "airflow" in k.lower()


def print_dbnd_environment_info():
    custom_print("DBND ENVIRONMENT INFO")
    custom_print("ENVIRON:\n %s" % pp.pformat(os.environ))
    custom_print(
        "DBND AIRFLOW ENVIRON:\n %s"
        % pp.pformat({k: v for k, v in os.environ.items() if _is_dbnd_airflow_key(k)})
    )

    custom_print("SYS PATH:\n %s" % pp.pformat(sys.path))
    custom_print("PYTHON:\n %s" % pp.pformat(sys.executable))


def print_dbnd_spark_info():
    custom_print("IS_DBND_SPARK_INSTALLED:  %s" % _is_dbnd_spark_installed())
    custom_print("VERIFY_SPARK_PRE_CONDITIONS: %s" % verify_spark_pre_conditions())
    print_airflow_context()

    from pyspark import SparkConf

    spark_conf = SparkConf().getAll()
    custom_print("DBND SPARK INFO")
    custom_print("SPARK GET CONF %s" % pp.pformat(spark_conf))
    custom_print(
        "DBND AIRFLOW SPARK GET CONF\n %s"
        % pp.pformat([(k, v) for k, v in spark_conf if _is_dbnd_airflow_key(k)])
    )


def print_airflow_context():
    custom_print("DBND SPARK AIRFLOW CONTEXT")
    try:
        af_context = try_get_airflow_context_from_spark_conf()
        if af_context:
            custom_print("DAG_ID: %s" % af_context.dag_id)
            custom_print("EXECUTION_DATE: %s" % af_context.execution_date)
            custom_print("TASK_ID: %s" % af_context.task_id)
            custom_print("TRY_NUMBER: %s" % af_context.try_number)
            custom_print("AIRFLOW_INSTANCE_UID: %s" % af_context.airflow_instance_uid)
        else:
            custom_print("DBND AIRFLOW CONTEXT is None")
    except Exception as e:
        custom_print("airflow context exception: %s" % pp.pformat(e))


def print_dbnd_debug():
    try:
        print_dbnd_environment_info()
        print_dbnd_spark_info()
    except Exception as e:
        custom_print("debug spark init exception: %s" % pp.pformat(e))


def create_dummy_dataframe(spark):
    try:
        from pyspark.sql.types import IntegerType, StringType, StructField, StructType

        data = [
            ("James", "", "Smith", "36636", "M", 3000),
            ("Michael", "Rose", "", "40288", "M", 4000),
            ("Robert", "", "Williams", "42114", "M", 4000),
            ("Maria", "Anne", "Jones", "39192", "F", 4000),
            ("Jen", "Mary", "Brown", "", "F", -1),
        ]

        schema = StructType(
            [
                StructField("firstname", StringType(), True),
                StructField("middlename", StringType(), True),
                StructField("lastname", StringType(), True),
                StructField("id", StringType(), True),
                StructField("gender", StringType(), True),
                StructField("salary", IntegerType(), True),
            ]
        )
        df = spark.createDataFrame(data=data, schema=schema)
        return df
    except Exception as e:
        custom_print("create dummy dataframe exception: %s" % pp.pformat(e))


def debug_log_dataset_op_read(df):
    from dbnd import dataset_op_logger

    with dataset_op_logger("dummy_df", "read", with_stats=True) as logger:
        logger.set(data=df)


def debug_log_dataset_op_write(df):
    from dbnd import dataset_op_logger

    with dataset_op_logger("dummy_df", "write", with_stats=True) as logger:
        df.write.format("csv").mode("overwrite").save("dummy_df.csv")
        logger.set(data=df)


def dbnd_log_debug_dataframe(df):
    if df:
        data_schema = {
            "columns": list(df.schema.names),
            "dtypes": {f.name: str(f.dataType) for f in df.schema.fields},
        }

        rows = df.count()
        data_dimensions = (rows, len(df.columns))
        data_schema.update(
            {"size.bytes": int(rows * len(df.columns)), "shape": data_dimensions}
        )
        custom_print("DATAFRAME SCHEMA:\n %s" % pp.pformat(data_schema))
    else:
        custom_print("DATAFRAME is None")


def dbnd_debug_spark_operations(spark):
    try:
        custom_print("dbnd debug spark operations")
        df = create_dummy_dataframe(spark)
        dbnd_log_debug_dataframe(df)
        debug_log_dataset_op_read(df)
        debug_log_dataset_op_write(df)
    except Exception as e:
        custom_print("dbnd debug spark operations exception: %s" % pp.pformat(e))


def dbnd_log_debug_spark():
    # print debug vars before calling dbnd_tracking_start
    print_dbnd_debug()

    # Try to "manually" start dbnd tracking
    try:
        from dbnd import dbnd_tracking_start

        task_run = dbnd_tracking_start()
        if task_run:
            custom_print("DBND_TRACKING_TASK_RUN: %s" % task_run)
            custom_print("DBND_TRACKING_RUN_UID: %s" % task_run.run.run_uid)
    except Exception as e:
        custom_print("dbnd tracking task run exception: %s" % e)

    # test log metric
    from dbnd import log_metric

    log_metric("debug.metric", 10000)

    # print debug vars after calling dbnd_tracking_start
    print_dbnd_debug()



# © Copyright Databand.ai, an IBM Company 2022

from datetime import datetime
from typing import List, Optional
from uuid import UUID

import attr

from dbnd._core.tracking.schemas.base import ApiStrictSchema
from dbnd._core.utils.basics.nothing import NOTHING
from dbnd._vendor.marshmallow import fields, post_load
from targets.data_schema import DataSchemaArgs, StructuredDataSchema


class GetRunningDagRunsResponseSchema(ApiStrictSchema):
    dag_run_ids = fields.List(fields.Integer())
    last_seen_dag_run_id = fields.Integer(allow_none=True)
    last_seen_log_id = fields.Integer(allow_none=True)


class UpdateLastSeenValuesRequestSchema(ApiStrictSchema):
    last_seen_dag_run_id = fields.Integer()
    last_seen_log_id = fields.Integer()


class UpdateAirflowMonitorStateRequestSchema(ApiStrictSchema):
    airflow_version = fields.String(required=False, allow_none=True)
    airflow_export_version = fields.String(required=False, allow_none=True)
    airflow_monitor_version = fields.String(required=False, allow_none=True)
    monitor_status = fields.String(required=False, allow_none=True)
    monitor_error_message = fields.String(required=False, allow_none=True)
    airflow_instance_uid = fields.UUID(required=False, allow_none=True)
    api_mode = fields.String(required=False, allow_none=True)


class GetAllDagRunsRequestSchema(ApiStrictSchema):
    min_start_time = fields.DateTime(allow_none=True)
    dag_ids = fields.String(allow_none=True)


class TaskSchema(ApiStrictSchema):
    upstream_task_ids = fields.List(fields.String())
    downstream_task_ids = fields.List(fields.String())
    task_type = fields.String()
    task_source_code = fields.String(allow_none=True)
    task_source_hash = fields.String(allow_none=True)
    task_module_code = fields.String(allow_none=True)
    module_source_hash = fields.String(allow_none=True)
    dag_id = fields.String()
    task_id = fields.String()
    retries = fields.Integer()
    command = fields.String(allow_none=True)
    task_args = fields.Dict()


class DagSchema(ApiStrictSchema):
    description = fields.String()
    root_task_ids = fields.List(fields.String())
    tasks = fields.Nested(TaskSchema, many=True)
    owner = fields.String()
    dag_id = fields.String()
    schedule_interval = fields.String()
    catchup = fields.Boolean(allow_none=True)
    start_date = fields.DateTime(allow_none=True)
    end_date = fields.DateTime(allow_none=True)
    is_committed = fields.Boolean()
    git_commit = fields.String()
    dag_folder = fields.String()
    hostname = fields.String()
    source_code = fields.String(allow_none=True)
    module_source_hash = fields.String(allow_none=True)
    tasks_hash_to_source = fields.Dict(allow_none=True)
    is_subdag = fields.Boolean()
    tags = fields.List(fields.String(), allow_none=True)
    task_type = fields.String()
    task_args = fields.Dict()
    is_active = fields.Boolean(allow_none=True)
    is_paused = fields.Boolean(allow_none=True)


class DagRunSchema(ApiStrictSchema):
    dag_id = fields.String(allow_none=True)
    run_id = fields.String(required=False)
    dagrun_id = fields.Integer()
    start_date = fields.DateTime(allow_none=True)
    state = fields.String()
    end_date = fields.DateTime(allow_none=True)
    execution_date = fields.DateTime(allow_none=True)
    task_args = fields.Dict()


class TaskInstanceSchema(ApiStrictSchema):
    execution_date = fields.DateTime()
    dag_id = fields.String()
    state = fields.String(allow_none=True)
    try_number = fields.Integer()
    task_id = fields.String()
    start_date = fields.DateTime(allow_none=True)
    end_date = fields.DateTime(allow_none=True)
    log_body = fields.String(allow_none=True)
    xcom_dict = fields.Dict()


class MetricsSchema(ApiStrictSchema):
    performance = fields.Dict()
    sizes = fields.Dict()


class AirflowExportMetaSchema(ApiStrictSchema):
    airflow_version = fields.String()
    plugin_version = fields.String()
    request_args = fields.Dict()
    metrics = fields.Nested(MetricsSchema)


class InitDagRunsRequestSchema(ApiStrictSchema):
    dags = fields.Nested(DagSchema, many=True)
    dag_runs = fields.Nested(DagRunSchema, many=True)
    task_instances = fields.Nested(TaskInstanceSchema, many=True)

    airflow_export_meta = fields.Nested(AirflowExportMetaSchema, required=False)
    error_message = fields.String(required=False, allow_none=True)
    syncer_type = fields.String(allow_none=True)


class OkResponseSchema(ApiStrictSchema):
    ok = fields.Boolean()


class UpdateDagRunsRequestSchema(ApiStrictSchema):
    dag_runs = fields.Nested(DagRunSchema, many=True)
    task_instances = fields.Nested(TaskInstanceSchema, many=True)
    last_seen_log_id = fields.Integer(allow_none=True)

    airflow_export_meta = fields.Nested(AirflowExportMetaSchema, required=False)
    error_message = fields.String(required=False, allow_none=True)
    syncer_type = fields.String(allow_none=True)


@attr.s
class BaseSourceMonitorState(object):
    monitor_version = attr.ib(default=NOTHING)  # type: str
    monitor_status = attr.ib(default=NOTHING)  # type: str
    monitor_error_message = attr.ib(default=NOTHING)  # type: str
    monitor_start_time = attr.ib(default=NOTHING)  # type: datetime

    def as_dict(self):
        # don't serialize data which didn't changed: as_dict should be able to return
        # None value when it set, specifically for monitor_error_message - when not set
        # at all (=NOTHING, didn't changed, =None changed)
        return attr.asdict(self, filter=lambda field, value: value is not NOTHING)


class BaseSourceMonitorStateSchema(ApiStrictSchema):
    monitor_version = fields.String(required=False, allow_none=True)
    monitor_status = fields.String(required=False, allow_none=True)
    monitor_error_message = fields.String(required=False, allow_none=True)
    monitor_start_time = fields.DateTime(required=False, allow_none=True)

    @post_load
    def make_object(self, data):
        return BaseSourceMonitorState(**data)


# ----- Datasource Tracking ------

# Datasource Monitor
@attr.s
class DatasourceMonitorState(object):
    datasource_monitor_version = attr.ib(default=NOTHING)  # type: str
    monitor_status = attr.ib(default=NOTHING)  # type: str
    monitor_error_message = attr.ib(default=NOTHING)  # type: str
    monitor_start_time = attr.ib(default=NOTHING)  # type: datetime
    last_sync_time = attr.ib(default=NOTHING)  # type: datetime
    last_transaction_sync_time = attr.ib(default=NOTHING)  # type: datetime

    def as_dict(self):
        # don't serialize data which didn't changed: as_dict should be able to return
        # None value when it set, specifically for monitor_error_message - when not set
        # at all (=NOTHING, didn't changed, =None changed)
        return attr.asdict(self, filter=lambda field, value: value is not NOTHING)


class DatasourceMonitorStateSchema(ApiStrictSchema):
    datasource_monitor_version = fields.String(required=False, allow_none=True)
    monitor_status = fields.String(required=False, allow_none=True)
    monitor_error_message = fields.String(required=False, allow_none=True)
    monitor_start_time = fields.DateTime(required=False, allow_none=True)
    last_sync_time = fields.DateTime(required=False, allow_none=True)
    last_transaction_sync_time = fields.DateTime(required=False, allow_none=True)

    @post_load
    def make_object(self, data):
        return DatasourceMonitorState(**data)


# Dataset Objects
@attr.s
class SyncedDatasetMetadata(object):
    num_bytes = attr.ib()  # type: int
    records = attr.ib()  # type: int
    schema = attr.ib()  # type: DataSchemaArgs

    def as_dict(self):
        return attr.asdict(self)


@attr.s
class SyncedDataset(object):
    uri = attr.ib()  # type: str
    created_date = attr.ib()  # type: datetime
    last_modified_date = attr.ib()  # type: datetime
    metadata = attr.ib()  # type: SyncedDatasetMetadata

    uid = attr.ib(default=None)  # type: UUID

    def as_dict(self):
        return attr.asdict(self)


@attr.s
class DatasetsReport(object):
    sync_event_uid = attr.ib()  # type: UUID
    monitor_state = attr.ib()  # type: DatasourceMonitorState

    sync_event_timestamp = attr.ib()  # type: datetime
    datasets = attr.ib()  # type: List[SyncedDataset]

    source_type = attr.ib(default=None)  # type: Optional[str]
    syncer_type = attr.ib(default=None)  # type: Optional[str]

    def as_dict(self):
        return attr.asdict(self, filter=lambda field, value: value is not NOTHING)


# Dataset schemas
class SyncedDatasetMetadataSchema(ApiStrictSchema):
    num_bytes = fields.Integer()
    records = fields.Integer()

    schema = fields.Nested(StructuredDataSchema)

    # TODO: ADD -> Preview

    @post_load
    def make_object(self, data):
        return SyncedDatasetMetadata(**data)


class SyncedDatasetSchema(ApiStrictSchema):
    uri = fields.String()  # {storage_type://region/project_id/scheme_name/table_name}
    created_date = fields.DateTime()
    last_modified_date = fields.DateTime(allow_none=True)

    metadata = fields.Nested(SyncedDatasetMetadataSchema)

    uid = fields.UUID(allow_none=True)

    @post_load
    def make_object(self, data):
        return SyncedDataset(**data)


class DatasetsReportSchema(ApiStrictSchema):
    sync_event_uid = fields.UUID()
    monitor_state = fields.Nested(DatasourceMonitorStateSchema)
    source_type = fields.String(allow_none=True)  # bigquery / snowflake / etc
    syncer_type = fields.String(allow_none=True)

    sync_event_timestamp = fields.DateTime(required=False, allow_none=True)
    datasets = fields.Nested(SyncedDatasetSchema, many=True)

    @post_load
    def make_object(self, data):
        return DatasetsReport(**data)


@attr.s
class SyncedTransactionOperation(object):
    op_type = attr.ib()  # type: str
    records_count = attr.ib()  # type: int
    started_date = attr.ib()  # type: datetime

    dataset_uri = attr.ib()  # type: str
    dataset_uid = attr.ib(default=None)  # type: UUID

    def as_dict(self, fields_to_ignore=None):
        fields_to_ignore = fields_to_ignore or []
        return attr.asdict(
            self, filter=lambda field, _: field.name not in fields_to_ignore
        )


@attr.s
class SyncedTransaction(object):
    datasource_transaction_id = attr.ib()  # type: str

    created_date = attr.ib()  # type: datetime
    started_date = attr.ib()  # type: datetime
    ended_date = attr.ib()  # type: datetime

    write_operation = attr.ib(
        default=None
    )  # type: Optional[SyncedTransactionOperation]
    read_operations = attr.ib(
        default=attr.Factory(list)
    )  # type: List[SyncedTransactionOperation]
    query_string = attr.ib(default=None)  # type: Optional[str]
    data_schema = attr.ib(default=None)  # type: Optional[DataSchemaArgs]

    def as_dict(self):
        return attr.asdict(self)

    @property
    def all_operations(self):
        # transaction can have multiple read operations but single write operation
        all_ops = self.read_operations.copy()

        if self.write_operation:
            all_ops.append(self.write_operation)

        return all_ops


@attr.s
class SyncedTransactionsReport(object):
    sync_event_uid = attr.ib()  # type: UUID
    sync_event_timestamp = attr.ib()  # type: datetime

    monitor_state = attr.ib()  # type: DatasourceMonitorStateSchema
    transactions = attr.ib()  # type: List[SyncedTransaction]

    source_type = attr.ib(default=None)  # type: str
    syncer_type = attr.ib(default=None)  # type: str

    def as_dict(self):
        return attr.asdict(self, filter=lambda field, value: value is not NOTHING)


class SyncedTransactionOperationSchema(ApiStrictSchema):
    op_type = fields.String()
    started_date = fields.DateTime()  # type: datetime
    records_count = fields.Integer()

    dataset_uri = fields.String()
    dataset_uid = fields.UUID(allow_none=True)

    @post_load
    def make_object(self, data):
        return SyncedTransactionOperation(**data)


class SyncedTransactionSchema(ApiStrictSchema):
    datasource_transaction_id = fields.String()
    created_date = fields.DateTime()
    started_date = fields.DateTime()
    ended_date = fields.DateTime()

    write_operation = fields.Nested(SyncedTransactionOperationSchema)
    read_operations = fields.Nested(
        SyncedTransactionOperationSchema, many=True, allow_none=True
    )

    query_string = fields.String(allow_none=True)
    data_schema = fields.Nested(StructuredDataSchema, allow_none=True)

    @post_load
    def make_object(self, data):
        return SyncedTransaction(**data)


class SyncedTransactionsReportSchema(ApiStrictSchema):
    sync_event_uid = fields.UUID()
    sync_event_timestamp = fields.DateTime()
    source_type = fields.String(allow_none=True)  # bigquery / snowflake / etc
    syncer_type = fields.String(allow_none=True)

    monitor_state = fields.Nested(DatasourceMonitorStateSchema)
    transactions = fields.Nested(SyncedTransactionSchema, many=True)

    @post_load
    def make_object(self, data):
        return SyncedTransactionsReport(**data)



# © Copyright Databand.ai, an IBM Company 2022

import json
import logging


logger = logging.getLogger(__name__)


class DatabandError(Exception):
    """
    Base exception.
    """

    _default_show_exc_info = True

    def __init__(self, *args, **kwargs):
        help_msg = kwargs.pop("help_msg", None)
        nested_exceptions = kwargs.pop("nested_exceptions", None)
        show_exc_info = kwargs.pop("show_exc_info", self._default_show_exc_info)

        super(DatabandError, self).__init__(*args, **kwargs)

        self.help_msg = help_msg
        self.show_exc_info = show_exc_info

        from dbnd._core.errors.errors_utils import UserCodeDetector

        self.user_frame_info = (
            UserCodeDetector.build_code_detector().find_user_side_frame(
                user_side_only=True
            )
        )

        if nested_exceptions:
            if not isinstance(nested_exceptions, list):
                nested_exceptions = [nested_exceptions]
        else:
            nested_exceptions = []

        try:
            self.nested_exceptions = nested_exceptions
            from dbnd._core.configuration.environ_config import is_unit_test_mode

            if is_unit_test_mode():
                from dbnd._core.errors.errors_utils import nested_exceptions_str

                if nested_exceptions:
                    logger.exception(
                        "PYTEST INFO: There are nested exceptions : %s",
                        nested_exceptions_str(self),
                    )
        except Exception:
            logger.exception("Failed to print nested exceptions for unit test mode")


class WrapperDatabandError(DatabandError):
    """
    Implement the boilerplate required to have a DatabandError which wrapping another DatabandError.
    Use this when you want to create a DatabandError that can chain inner error inside but keep the properties
     of the inner error.
    """

    def __init__(self, message, inner_error, **kwargs):
        # type: (str, Exception, dict) -> WrapperDatabandError
        extended_msg = (
            "{message}, caused by: \n\t {cause_name}: {cause_message} ".format(
                message=message,
                cause_name=inner_error.__class__.__name__,
                cause_message=str(inner_error),
            )
        )
        super(WrapperDatabandError, self).__init__(extended_msg, **kwargs)
        self.inner_error = inner_error

        # try passing the help message of the cause to this error if current error is None
        try:
            self.help_msg = self.help_msg or self.inner_error.help_msg
        except AttributeError:
            pass


class DatabandBuildError(DatabandError):
    """
    Error while building task
    """

    _default_show_exc_info = False


class DatabandRuntimeError(DatabandError):
    """
    Error in task run
    """

    _default_show_exc_info = False


class DatabandSigTermError(DatabandError):
    """
    SIGTERM received by the process
    """

    _default_show_exc_info = False


class DatabandSystemError(Exception):
    """
    Databand error that should not happen!
    """

    _default_show_exc_info = True


class DatabandConfigError(DatabandError):
    _default_show_exc_info = False


class DatabandRunError(DatabandError):
    """
    Databand executor(airflow) failed to run some tasks
    """

    _default_show_exc_info = False

    def __init__(self, *args, **kwargs):
        run = kwargs.pop("run", None)

        super(DatabandRunError, self).__init__(*args, **kwargs)
        if run is None:
            from dbnd import get_databand_run

            run = get_databand_run()

        self.run = run


class DatabandFailFastError(DatabandRunError):
    """
    Databand executor has terminated because a task failed
    """


class ConfigLookupError(DatabandConfigError):
    """
    Error signifying that the wanted configuration wasn't found
    """

    _default_show_exc_info = False


class ParameterError(DatabandConfigError):
    """
    Base exception.
    """

    _default_show_exc_info = False


class ParseValueError(ParameterError):
    """
    Error signifying that there was a error whily parsing Parameter.
    """


class ValueTypeError(DatabandConfigError):
    """
    Base exception.
    """

    _default_show_exc_info = False


class ParseParameterError(ParameterError):
    """
    Error signifying that there was a error whily parsing Parameter.
    """


class MissingParameterError(ParameterError):
    """
    Error signifying that there was a missing Parameter.
    """


class UnknownParameterError(ParameterError):
    """
    Error signifying that an unknown Parameter was supplied.
    """


class DbndCanceledRunError(DatabandError):
    """
    Error thrown when canceling a run.
    """


class TaskClassException(DatabandError):
    _default_show_exc_info = False


class TaskClassNotFoundException(TaskClassException):
    pass


class TaskClassAmbigiousException(TaskClassException):
    pass


class TaskValidationError(DatabandError):
    pass


class DatabandApiError(DatabandError):
    def __init__(self, method, endpoint, resp_code, response, **kwargs):
        super(DatabandApiError, self).__init__(
            method, endpoint, resp_code, response, **kwargs
        )

        self.method = method
        self.endpoint = endpoint
        self.resp_code = resp_code
        self.response = response

    def __str__(self):
        try:
            parsed_json = json.loads(self.response)
            error_message = json.dumps(parsed_json, indent=4)
        except Exception:
            error_message = self.response

        return "Call failed to endpoint %s %s\nResponse code: %s\nServer error: %s" % (
            self.method,
            self.endpoint,
            self.resp_code,
            error_message.replace("\n", "\n\t"),
        )


class DatabandUnauthorizedApiError(DatabandApiError):
    """Api error indicate unauthorized http request (status code: 403 or 401)"""


class DatabandBadRequest(DatabandError):
    pass


class DatabandWebserverNotReachableError(DatabandError):
    """
    Indicate that Databand webserver is not reachable
    """


class DatabandAuthenticationError(DatabandWebserverNotReachableError):
    """Api error indicate we failed to authenticate to the server"""


class DatabandConnectionException(DatabandWebserverNotReachableError):
    """
    Error thrown when connecting with the server is not available.
    """



# © Copyright Databand.ai, an IBM Company 2022

from os import path

import setuptools

from setuptools.config import read_configuration


BASE_PATH = path.dirname(__file__)
CFG_PATH = path.join(BASE_PATH, "setup.cfg")

config = read_configuration(CFG_PATH)

# A list of vendored packages
dbnd_vendors_list = [
    "cachetools",
    "hjson",
    "cloudpickle",
    "pendulum==1.4.4",
    "tabulate",
    "marshmallow==2.18.0",
    "croniter>=0.3.30,<0.4",
    "psutil>=4.2.0,<5.7.0",  # extracted use to vendorized_psutil.py
]

setuptools.setup(
    name="dbnd",
    package_dir={"": "src"},
    python_requires=">=3.6, <=3.10",
    install_requires=[
        "tzlocal",
        "six",
        "more-itertools",
        "attrs!=21.1.0",  # yanked version, breaks dbnd
        "pyyaml",  # yaml support in targets
        "pytz",  # python time zone for pendulum library
        "pytzdata",  # python time zone for pendulum library
        "requests>=2.18.0",  # API TRACKING
        "configparser<3.6.0,>=3.5.0",  # same versions as Airflow -- Moved library to _vendor
        "pygments>=2.6.1",
        # backward compatible python
        'typing;python_version<"3.7"',  # Standalone pkg is incompatible with 3.7 and not required
        "python-dateutil",
        "sqlparse",
        # WARNING! Keep install_requires in sync with dbnd.requirements.txt file:
        # use dbnd.requirements.txt file generated by `make dist-python`
    ],
    extras_require={
        ':sys_platform=="win32"': ["colorama"],
        "tests": [
            "qtconsole==4.7.7",
            "numpy<1.23",
            "coverage",
            "pytest==4.5.0",  # 4.6.0 requires pluggy 0.12
            "pytest-cov==2.9.0",
            "pluggy==0.11.0",  # 0.12 has import_metadata, fails on py2
            "zope.interface",
            "mock",
            "pandas<2.0.0,>=0.17.1",  # airflow supports only this version
            "tox==3.12.1",
            "matplotlib==3.3.0",
            "tables==3.7.0",
            "feather-format",
            "pyarrow",
            # conflict with pandas version on new openpyxl: got invalid input value of type <class 'xml.etree.ElementTree.Element'>, expected string or Element
            "openpyxl==2.6.4",
            "scikit-learn",
            "wheel",  # for fat_wheel tests
        ],
        "jupyter": [
            "nbconvert",
            "nbformat",
            "jupyter",
            "traitlets>=4.2,<5.0.0",  # required by jupyter, fix py37 compatibility
            "IPython",
            "jupyter_contrib_nbextensions",
        ],
    },
    entry_points={"console_scripts": ["dbnd = dbnd:dbnd_main"]},
)



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.task_build.task_signature import TASK_ID_INVALID_CHAR_REGEX
from targets import target


class TaskRunMetaFiles(object):
    def __init__(self, root):
        self.root = root

    _ARTIFACTS = "artifacts"
    _METRICS = "metrics"
    _META_DATA_FILE_NAME = "meta.yaml"
    _DEFAULT_METRIC_SOURCE = "user"

    def _output(self, *path):
        return target(self.root, *path)

    def get_metric_folder(self, source=None):
        source = source or TaskRunMetaFiles._DEFAULT_METRIC_SOURCE
        return self._output(TaskRunMetaFiles._METRICS, source, "")  # type: DirTarget

    def get_metric_target(self, metric_key, source=None):
        metric_key = TASK_ID_INVALID_CHAR_REGEX.sub("_", metric_key)
        source = source or TaskRunMetaFiles._DEFAULT_METRIC_SOURCE
        return self._output(TaskRunMetaFiles._METRICS, source, metric_key)

    def get_artifact_target(self, name):
        return self._output(TaskRunMetaFiles._ARTIFACTS, name)

    def get_meta_data_file(self):
        return self._output(TaskRunMetaFiles._META_DATA_FILE_NAME)



# © Copyright Databand.ai, an IBM Company 2022

import logging
import os
import typing

from typing import Any, ContextManager, List, Optional, Union
from uuid import UUID

from dbnd._core.configuration.environ_config import (
    DBND_PARENT_TASK_RUN_ATTEMPT_UID,
    DBND_PARENT_TASK_RUN_UID,
    DBND_ROOT_RUN_TRACKER_URL,
    DBND_ROOT_RUN_UID,
    DBND_RUN_UID,
    ENV_DBND__USER_PRE_INIT,
)
from dbnd._core.constants import (
    AD_HOC_DAG_PREFIX,
    RESULT_PARAM,
    SystemTaskName,
    TaskRunState,
    UpdateSource,
)
from dbnd._core.errors import DatabandRuntimeError
from dbnd._core.run.run_banner import RunBanner
from dbnd._core.run.run_tracker import RunTracker
from dbnd._core.run.target_identity_source_map import TargetIdentitySourceMap
from dbnd._core.settings import DatabandSettings
from dbnd._core.settings.engine import build_engine_config
from dbnd._core.task import Task
from dbnd._core.task_build.task_context import current_task, has_current_task
from dbnd._core.task_run.task_run import TaskRun
from dbnd._core.tracking.airflow_dag_inplace_tracking import AirflowTaskContext
from dbnd._core.tracking.schemas.tracking_info_run import RootRunInfo, ScheduledRunInfo
from dbnd._core.utils.basics.singleton_context import SingletonContext
from dbnd._core.utils.date_utils import unique_execution_date
from dbnd._core.utils.traversing import flatten
from dbnd._core.utils.uid_utils import get_uuid
from dbnd._vendor.namesgenerator import get_random_name
from targets.caching import TARGET_CACHE


if typing.TYPE_CHECKING:
    from dbnd._core.context.databand_context import DatabandContext
    from dbnd._core.task_executor.run_executor import RunExecutor

logger = logging.getLogger(__name__)


class DatabandRun(SingletonContext):
    """
    Main object that represents Pipeline/Tasks/Flow Run
    it controls
     * the overall state of the execution
     * tracking of the whole run
     * tracking of the data objects between different tasks
    Main two modes:
     * tracking - driver is optional, root task has to be assigned
     * orchestration - driver will be assigned by RunExecutor
    """

    def __init__(
        self,
        context,  # type: DatabandContext
        job_name,
        project_name=None,  # type: Optional[str]
        run_uid=None,  # type:  Optional[UUID]
        scheduled_run_info=None,  # type:  Optional[ScheduledRunInfo]
        existing_run=False,  # type:  Optional[bool]
        source=UpdateSource.dbnd,  # type:  Optional[UpdateSource]
        af_context=None,
        tracking_source=None,
        is_orchestration=False,
    ):
        self.context = context
        s = self.context.settings  # type: DatabandSettings

        self.job_name = job_name
        self.project_name = project_name or s.tracking.project

        self.description = s.run.description
        self.is_archived = s.run.is_archived
        self.source = source
        self.is_orchestration = is_orchestration

        self.existing_run = existing_run or False
        # this was added to allow the scheduler to create the run which will be continued by the actually run command instead of having 2 separate runs
        if not run_uid and DBND_RUN_UID in os.environ:
            # we pop so if this run spawnes subprocesses with their own runs they will be associated using the sub-runs mechanism instead
            # of being fused into this run directly
            run_uid = UUID(os.environ.pop(DBND_RUN_UID))
            self.existing_run = True

        self.run_uid = run_uid or get_uuid()
        # if user provided name - use it
        # otherwise - generate a name for the run
        self.name = s.run.name or self._generate_run_name(af_context)
        self.execution_date = (
            self.context.settings.run.execution_date or unique_execution_date()
        )

        self.is_tracked = True

        # tracking/orchestration main task
        self.root_task = None  # type: Optional[Task]

        # task run that wraps execution (tracking or orchestration)
        self._driver_task_run = None

        # ORCHESTRATION: execution of the run
        self.run_executor = None  # type: Optional[RunExecutor]

        # dag_id , execution_date are used by Airflow,
        # should be deprecated (still used by DB tracking)
        self.dag_id = (
            (AD_HOC_DAG_PREFIX + self.job_name) if is_orchestration else self.job_name
        )

        # RUN STATE
        self._run_state = None
        self.task_runs = []  # type: List[TaskRun]
        self.task_runs_by_id = {}
        self.task_runs_by_af_id = {}

        self.target_origin = TargetIdentitySourceMap()
        self.describe = RunBanner(self)
        self.tracker = RunTracker(self, tracking_store=self.context.tracking_store)

        # ALL RUN CONTEXT SPECIFIC thing
        self.root_run_info = RootRunInfo.from_env(current_run=self)
        self.scheduled_run_info = scheduled_run_info or ScheduledRunInfo.from_env(
            self.run_uid
        )
        self.env = self.context.env
        self.run_folder_prefix = os.path.join(
            "log",
            self.execution_date.strftime("%Y-%m-%d"),
            "%s_%s_%s"
            % (
                self.execution_date.strftime("%Y-%m-%dT%H%M%S.%f"),
                self.job_name,
                self.name,
            ),
        )
        self.run_root = self.env.dbnd_root.folder(self.run_folder_prefix)
        self.run_local_root = self.env.dbnd_local_root.folder(self.run_folder_prefix)

        self.local_engine = build_engine_config(self.env.local_engine).clone(
            require_submit=False
        )

        self.dynamic_af_tasks_count = dict()
        self.af_context = af_context
        self.tracking_source = tracking_source
        self.start_time = None
        self.finished_time = None
        self._result_location = None

    def _generate_run_name(self, af_context: Optional[AirflowTaskContext]) -> str:
        """
        If this is an airflow run, generate a name to reflect it still
        awaits a sync.
        Otherwise, generate a human friendly name for the run.
        """
        if af_context is not None:
            return f"Airflow-run-await-sync_{self.run_uid}"
        return get_random_name(seed=self.run_uid)

    def get_task_runs(self, without_executor=True, without_system=False):
        """
        :param without_executor: filter driver/submitter task runs
        :param without_system:   filter task.is_system tasks
        :return: List[TaskRun]
        """
        task_runs = self.task_runs
        if without_executor:
            task_runs = [
                tr
                for tr in task_runs
                if tr.task.task_name not in SystemTaskName.driver_and_submitter
            ]
        if without_system:
            task_runs = [tr for tr in task_runs if not tr.task.task_is_system]
        return task_runs

    @property
    def root_task_run(self):  # type: (...)->  Optional[TaskRun]
        if self.root_task:
            return self.get_task_run_by_id(self.root_task.task_id)
        return None

    @property
    def driver_task_run(self):
        """
        This is the main "wrapper" task that will have the "log" of the code execution
        :return:
        """
        # in case orchestration/luigi/others has explicit task run
        if self._driver_task_run:
            return self._driver_task_run
        # otherwise we should use just a main task as a "driver" task (tracking case)
        return self.root_task_run

    @driver_task_run.setter
    def driver_task_run(self, value):
        self._driver_task_run = value

    def build_and_set_driver_task_run(self, driver_task, driver_engine=None):
        """
        set driver task run which is used to "track" main execution flow
        Tracking: will track pipeline progress
        Orchestration: will track and "execute" pipeline

        you need to run DatabandRun.init_run otherwise it will be not tracked
        """
        self._driver_task_run = TaskRun(
            task=driver_task, run=self, task_engine=driver_engine or self.local_engine
        )
        self._add_task_run(self._driver_task_run)
        return self._driver_task_run

    @property
    def run_url(self):
        return self.tracker.run_url

    @property
    def task(self):
        return self.root_task

    # TODO: split to get_by_id/by_af_id
    def get_task_run(self, task_id):
        # type: (str) -> TaskRun
        return self.get_task_run_by_id(task_id) or self.get_task_run_by_af_id(task_id)

    def get_task_run_by_id(self, task_id):
        # type: (str) -> TaskRun
        return self.task_runs_by_id.get(task_id)

    def get_task_run_by_af_id(self, task_id):
        # type: (str) -> TaskRun
        return self.task_runs_by_af_id.get(task_id)

    def get_af_task_ids(self, task_ids):
        return [self.get_task_run(task_id).task_af_id for task_id in task_ids]

    def get_task(self, task_id):
        # type: (str) -> Task
        return self.get_task_run(task_id).task

    @property
    def describe_dag(self):
        return self.root_task.ctrl.describe_dag

    def set_run_state(self, state):
        self._run_state = state
        self.tracker.set_run_state(state)

    @property
    def duration(self):
        if self.finished_time and self.start_time:
            return self.finished_time - self.start_time
        return None

    def _get_task_by_id(self, task_id):
        task = self.context.task_instance_cache.get_task_by_id(task_id)
        if task is None:
            raise DatabandRuntimeError(
                "Failed to find task %s in current context" % task_id
            )

        return task

    def next_af_task_name(self, task):
        """
        user friendly task name:
            task_HASH -> task
            task_HASH2 -> task_1
        :param task:
        :return:
        """
        task_name = task.friendly_task_name
        if task_name in self.dynamic_af_tasks_count:
            self.dynamic_af_tasks_count[task_name] += 1
            task_af_id = "{}_{}".format(
                task_name, self.dynamic_af_tasks_count[task_name]
            )
        else:
            self.dynamic_af_tasks_count[task_name] = 1
            task_af_id = task_name
        return task_af_id

    def create_task_run_at_execution_time(self, task, task_engine, task_af_id=None):
        if task_af_id is None:
            task_af_id = self.next_af_task_name(task)

        if self.af_context:
            task_af_id = "_".join([self.af_context.task_id, task_af_id])

        tr = TaskRun(
            task=task,
            run=self,
            is_dynamic=True,
            task_engine=task_engine,
            task_af_id=task_af_id,
        )
        self.add_task_runs_and_track([tr])
        return tr

    def add_task_runs_and_track(self, task_runs):
        # type: (List[TaskRun]) -> None
        for tr in task_runs:
            self._add_task_run(tr)

        self.tracker.add_task_runs(task_runs)

    def _build_and_add_task_run(
        self, task, task_engine=None, task_af_id=None, try_number=1
    ):
        if task_af_id is None:
            task_af_id = self.next_af_task_name(task)

        tr = TaskRun(
            task=task,
            run=self,
            task_engine=task_engine or self.local_engine,
            task_af_id=task_af_id,
            try_number=try_number,
        )
        self._add_task_run(tr)
        return tr

    def _add_task_run(self, task_run):
        self.task_runs.append(task_run)
        self.task_runs_by_id[task_run.task.task_id] = task_run
        self.task_runs_by_af_id[task_run.task_af_id] = task_run

        task_run.task.ctrl.last_task_run = task_run

    def cleanup_after_task_run(self, task):
        # type: (Task) -> None
        rels = task.ctrl.relations
        # potentially, all inputs/outputs targets for current task could be removed
        targets_to_clean = set(flatten([rels.task_inputs, rels.task_outputs]))

        targets_in_use = set()
        # any target which appears in inputs of all not finished tasks shouldn't be removed
        for tr in self.task_runs:
            if tr.task_run_state in TaskRunState.final_states():
                continue
            # remove all still needed inputs from targets_to_clean list
            for target in flatten(tr.task.ctrl.relations.task_inputs):
                targets_in_use.add(target)

        TARGET_CACHE.clear_for_targets(targets_to_clean - targets_in_use)

    def get_context_spawn_env(self):
        env = {}
        if has_current_task():
            current = current_task()
        else:
            current = self.root_task

        if current:
            tr = self.get_task_run_by_id(current.task_id)
            if tr:
                env[DBND_PARENT_TASK_RUN_UID] = str(tr.task_run_uid)
                env[DBND_PARENT_TASK_RUN_ATTEMPT_UID] = str(tr.task_run_attempt_uid)

        env[DBND_ROOT_RUN_UID] = str(self.root_run_info.root_run_uid)
        env[DBND_ROOT_RUN_TRACKER_URL] = self.root_run_info.root_run_url

        if self.context.settings.core.user_code_on_fork:
            env[ENV_DBND__USER_PRE_INIT] = self.context.settings.core.user_code_on_fork
        return env

    def is_killed(self):
        if self.run_executor:
            return self.run_executor.is_killed()
        return False

    def kill(self):
        """
        called from user space, kills the current task only
        :return:
        """
        if self.run_executor:
            self.run_executor.kill()

    def kill_run(self, message=None):
        if self.run_executor:
            self.run_executor.kill_run(message)

    def get_current_dbnd_local_root(self):
        # we should return here the proper engine config, based in which context we run right now
        # it could be submit, driver or task engine
        return self.env.dbnd_local_root

    def load_from_result(self, name=RESULT_PARAM, value_type=None):
        # type: (Optional[str], Optional[str]) -> Any
        """
        Loads the result of the run by name, default is the default result name.
        """
        return self.run_executor.result.load(name, value_type)

    def get_upstream_failed_task_run_state(self, task_run):
        # type: (TaskRun)-> TaskRunState
        """
        This function is a helper function to decide task final state only for orchestration runs.
        param: task_run - the task  which we want to get it's state
        return: TaskRunState
        """
        state = TaskRunState.UPSTREAM_FAILED
        child_task_runs = task_run.task.descendants
        # Since task run has children, IT'S a pipeline task, so we need to calculate it's state based on children states
        if child_task_runs:
            for task_run_id in child_task_runs.children:
                child_tr_instance = self.get_task_run_by_id(task_run_id)
                if child_tr_instance.task_run_state == TaskRunState.FAILED:
                    # In case one of its children fail,task run state is force updated from upstream failed to failed
                    #   because upstream_failed means task dependency is not met, while actually inner task has errors
                    state = TaskRunState.FAILED

        return state


def new_databand_run(context, job_name, run_uid=None, **kwargs):
    # type: (DatabandContext, Union[Task, str], UUID, **Any)-> ContextManager[DatabandRun]

    kwargs["allow_override"] = kwargs.pop("allow_override", True)
    return DatabandRun.new_context(
        context=context, job_name=job_name, run_uid=run_uid, **kwargs
    )



# Copyright (c) 2015  aggftw@gmail.com
# Distributed under the terms of the Modified BSD License.
# from sparkmagic package
NO_AUTH = "None"
AUTH_KERBEROS = "Kerberos"
AUTH_BASIC = "Basic_Access"
AUTHS_SUPPORTED = [NO_AUTH, AUTH_KERBEROS, AUTH_BASIC]


CONFIGURABLE_RETRY = "configurable"
LINEAR_RETRY = "linear"
LINEAR_RETRY_ANY_ERROR = "linear_retry_any_error"



# © Copyright Databand.ai, an IBM Company 2022

import typing

from collections.abc import Iterable

import attr


if typing.TYPE_CHECKING:
    from typing import Callable, List, Optional, Union


def columns_converter(columns):
    # type: (Union[str,Iterable[str], Callable[[], List[str]]]) -> List
    if isinstance(columns, str):
        return columns.split(",")
    elif isinstance(columns, Iterable):
        return [str(col) for col in columns]
    elif callable(columns):
        return columns()
    return []


@attr.s
class LogDataRequest(object):
    """
    Used in log_data() for with_stats, with_histograms params.

    Example::

        @task
        def process_customers_data(data):
            log_dataframe("customers_data", data,
                          with_histograms=LogDataRequest(include_all_numeric=True,
                                                           exclude_columns=["name", "phone"]))
    """

    include_columns = attr.ib(
        converter=columns_converter, factory=list
    )  # type: Union[Callable[[], List[str]], List]
    exclude_columns = attr.ib(
        converter=columns_converter, factory=list
    )  # type: Union[Callable[[], List[str]], List]
    include_all_boolean = attr.ib(default=False)  # type: bool
    include_all_numeric = attr.ib(default=False)  # type: bool
    include_all_string = attr.ib(default=False)  # type: bool

    # TODO: Make these helpers class properties
    @classmethod
    def ALL_BOOLEAN(cls):
        """Calculate statistics and histograms for all columns of boolean type(s)."""
        return cls(include_all_boolean=True)

    @classmethod
    def ALL_NUMERIC(cls):
        """Calculate statistics and histograms for all columns of Numeric type(s)."""
        return cls(include_all_numeric=True)

    @classmethod
    def ALL_STRING(cls):
        """
        Calculate statistics and histograms for all columns of String type(s).

        Usually this means all non boolean, non numeric types
        """
        return cls(include_all_string=True)

    @classmethod
    def ALL(cls):
        """Calculate statistics and histograms for all columns."""
        return cls(
            include_all_boolean=True, include_all_numeric=True, include_all_string=True
        )

    @classmethod
    def NONE(cls):
        return cls()

    def __bool__(self):
        return bool(
            self.include_columns
            or self.include_all_numeric
            or self.include_all_boolean
            or self.include_all_string
            or self.exclude_columns
        )

    __nonzero__ = __bool__

    @classmethod
    def from_user_param(cls, user_param):
        # type: (Optional[Union[bool, str, List[str], LogDataRequest]]) -> LogDataRequest
        """
        To allow flexibility for the user, the parameters can have other types.

        - boolean to calculate or not on all the data
        - list and str to specify specific column names
        """
        if isinstance(user_param, cls):
            return user_param
        elif isinstance(user_param, bool):
            return cls.ALL() if user_param else cls.NONE()
        elif isinstance(user_param, (list, str)):
            return cls(include_columns=user_param)



# © Copyright Databand.ai, an IBM Company 2022

from functools import reduce, wraps
from operator import xor

from dbnd._vendor import click


def only_one(*args):
    return reduce(xor, map(bool, args))


def required_mutually_exclusive_options(*options):
    """
    Creates a decorator for click commands that make sure that
    only one, and at least one, of `options` is set while calling the click command.
    It also set a relevant help message for the mutually excluded options.
    """

    def decorator(func):
        options_obj = [
            option for option in func.__click_params__ if option.name in options
        ]
        options_names = {option.name for option in options_obj}
        for option in options_obj:
            current_help = "" if option.help is None else option.help + " "
            others = list(options_names - {option.name})
            added_msg = "**mutually excluded with {others}".format(others=others)
            option.help = current_help + added_msg

        @wraps(func)
        def inner(**kwargs):
            relevant_options = [
                value for name, value in kwargs.items() if name in options
            ]
            if not only_one(*relevant_options):
                raise click.UsageError(
                    "Illegal usage: `at least one, and only one of the options [{options}] is required` ".format(
                        options=",".join("--" + op_name for op_name in options)
                    )
                )
            func(**kwargs)

        return inner

    return decorator


def options_dependency(root, *dependents):
    """
    Creates a decorator for click commands that make sure that only when `root`
    option is set all the others `dependents` could be used.
    It also set a relevant help message for the dependents options.
    """

    def decorator(func):
        options_obj = [
            option for option in func.__click_params__ if option.name in dependents
        ]
        for option in options_obj:
            current_help = "" if option.help is None else option.help + " "
            added_msg = "**depend on {root} option".format(root=root)
            option.help = current_help + added_msg

        @wraps(func)
        def inner(**kwargs):
            relevant_options = [
                value for name, value in kwargs.items() if name in dependents
            ]
            if any(relevant_options) and not kwargs.get(root):
                raise click.UsageError(
                    "Illegal usage: In-order to use any of those options {options}, "
                    "the `{root}` option must be set".format(
                        root=root, options=dependents
                    )
                )

            func(**kwargs)

        return inner

    return decorator



# © Copyright Databand.ai, an IBM Company 2022

import logging
import os
import threading
import typing

from dbnd._core.errors import DatabandRuntimeError
from dbnd._core.utils.seven import Callable
from targets.fs.local import LocalFileSystem


logger = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    from targets import FileSystem


class FileSystems(object):
    local = "local"
    s3 = "s3"
    gcs = "gs"


KNOWN_FILE_SYSTEMS = {}
CUSTOM_FILE_SYSTEM_MATCHERS = []


def register_file_system(prefix, file_system_factory):
    # type: (str,Callable[[], FileSystem]) -> None
    KNOWN_FILE_SYSTEMS[prefix] = file_system_factory


def register_file_system_name_custom_resolver(resolver):
    # type: (Callable[[str], str])->None
    CUSTOM_FILE_SYSTEM_MATCHERS.append(resolver)


# in memory is implemented as part of InMemory target
# register_file_system("memory", )
register_file_system(FileSystems.local, LocalFileSystem)
register_file_system("file", LocalFileSystem)

_cached_fs = {}


def reset_fs_cache():
    global _cached_fs
    _cached_fs = {}


def get_file_system(fs_name):
    current_thread_id = threading.current_thread().ident
    fs = _cached_fs.get((fs_name, current_thread_id))
    if fs:
        return fs

    if fs_name not in KNOWN_FILE_SYSTEMS:
        raise Exception("Unknown file system '%s' " % fs_name)

    fs_builder = KNOWN_FILE_SYSTEMS[fs_name]
    _cached_fs[(fs_name, current_thread_id)] = fs = fs_builder()
    return fs


def get_file_system_name(path):
    fs_prefix = None
    if ":" in path:
        fs_prefix = path.split(":")[0]
        if fs_prefix in KNOWN_FILE_SYSTEMS:
            return fs_prefix
        if os.name == "nt" and fs_prefix.lower() in get_windows_drives():
            return FileSystems.local

    for cfs in CUSTOM_FILE_SYSTEM_MATCHERS:
        found_fs_name = cfs(path)
        if found_fs_name:
            return found_fs_name

    if path.startswith("/"):
        return FileSystems.local

    if fs_prefix:
        # TODO: it would be nice to provide more useful information
        raise DatabandRuntimeError(
            "Can't find file system '%s'" % fs_prefix,
            help_msg="Please check that you have registered required schema with"
            " `register_file_system` or relevant plugin is installed",
        )
    return FileSystems.local


def get_windows_drives():
    import string

    from ctypes import windll

    drives = []
    bitmask = windll.kernel32.GetLogicalDrives()
    for letter in string.ascii_lowercase:
        if bitmask & 1:
            drives.append(letter)
        bitmask >>= 1

    return drives



# © Copyright Databand.ai, an IBM Company 2022

import datetime

from typing import Dict, Optional

import attr

from dbnd._core.errors.base import DatabandBadRequest
from dbnd._core.tracking.schemas.base import ApiStrictSchema
from dbnd._vendor.marshmallow import ValidationError, fields, post_load, validate


class MonitorConfigSchema(ApiStrictSchema):
    include_sources = fields.Boolean(required=False)
    dag_run_bulk_size = fields.Integer(required=False)
    start_time_window = fields.Integer(required=False)
    log_bytes_from_head = fields.Integer(required=False)
    log_bytes_from_end = fields.Integer(required=False)
    use_async_tracking = fields.Boolean(required=False)


class AirflowServerInfoSchema(ApiStrictSchema):
    airflow_version = fields.String(allow_none=True)
    airflow_export_version = fields.String(allow_none=True)
    airflow_monitor_version = fields.String(allow_none=True)
    last_sync_time = fields.DateTime(allow_none=True)
    monitor_error_message = fields.String(allow_none=True)
    synced_to = fields.DateTime(allow_none=True)
    api_mode = fields.String(allow_none=True)
    sync_interval = fields.Integer(allow_none=True)
    is_sync_enabled = fields.Boolean(allow_none=True)
    system_alert_definitions = fields.Dict()
    fetcher = fields.String(allow_none=True)
    composer_client_id = fields.String(allow_none=True)
    dag_ids = fields.String(allow_none=True)

    base_url = fields.String()
    external_url = fields.String(allow_none=True)
    source_instance_uid = fields.String(allow_none=True)
    tracking_source_uid = fields.String()
    airflow_instance_uid = fields.String(allow_none=True)  # TODO_API: deprecate
    name = fields.String(allow_none=True)
    env = fields.String(allow_none=True)
    monitor_status = fields.String(allow_none=True)
    monitor_config = fields.Nested(MonitorConfigSchema, allow_none=True)
    airflow_environment = fields.String(allow_none=True)
    last_seen_dag_run_id = fields.Integer(allow_none=True)
    last_seen_log_id = fields.Integer(allow_none=True)

    @post_load
    def make_object(self, data, **kwargs):
        return AirflowServerInfo(**data)


airflow_server_info_schema = AirflowServerInfoSchema()


def is_url_valid(url):
    return validate.URL(schemes={"http", "https"}, require_tld=False)(url)


def url_validation(obj, attribute, value):
    if value is not None and value:
        try:
            return is_url_valid(value)
        except ValidationError:
            raise DatabandBadRequest(
                "Could not edit airflow syncer because {} url {} was not a valid url".format(
                    attribute.name, value
                )
            )


@attr.s
class AirflowServerInfo(object):
    base_url = attr.ib(validator=url_validation)  # type: str
    monitor_status = attr.ib(default=None)  # type: Optional[str]
    name = attr.ib(default=None)  # type: Optional[str]
    env = attr.ib(default=None)  # type: Optional[str]
    external_url = attr.ib(
        default=None, validator=url_validation
    )  # type: Optional[str]
    source_instance_uid = attr.ib(default=None)  # type: str
    tracking_source_uid = attr.ib(default=None)  # type: str
    airflow_instance_uid = attr.ib(default=None)  # type: str #TODO_API: derperate
    airflow_version = attr.ib(default=None)  # type: Optional[str]
    airflow_export_version = attr.ib(default=None)  # type: Optional[str]
    airflow_monitor_version = attr.ib(default=None)  # type: Optional[str]
    last_sync_time = attr.ib(default=None)  # type: Optional[datetime.datetime]
    monitor_error_message = attr.ib(default=None)  # type: Optional[str]
    synced_to = attr.ib(default=None)  # type: Optional[datetime.datetime]
    api_mode = attr.ib(default="rbac")  # type: Optional[str]
    is_sync_enabled = attr.ib(default=None)  # type: Optional[bool]
    system_alert_definitions = attr.ib(default=None)  # type: Optional[Dict]
    fetcher = attr.ib(default=None)  # type: Optional[str]
    composer_client_id = attr.ib(default=None)  # type: Optional[str]
    dag_ids = attr.ib(default=None)  # type: Optional[str]
    monitor_config = attr.ib(default=None)  # type: Optional[Dict]
    airflow_environment = attr.ib(default=None)  # type: Optional[str]
    last_seen_dag_run_id = attr.ib(default=None)  # type: Optional[int]
    last_seen_log_id = attr.ib(default=None)  # type: Optional[int]



# © Copyright Databand.ai, an IBM Company 2022

import typing

from datetime import datetime
from typing import List, Optional, Tuple
from uuid import UUID

import attr

from dbnd._core.constants import (
    DbndDatasetOperationType,
    DbndTargetOperationStatus,
    DbndTargetOperationType,
    RunState,
    TaskRunState,
    UpdateSource,
)
from dbnd._core.tracking.airflow_dag_inplace_tracking import AirflowTaskContext
from dbnd._core.tracking.schemas.base import ApiStrictSchema
from dbnd._core.tracking.schemas.column_stats import ColumnStatsArgs, ColumnStatsSchema
from dbnd._core.tracking.schemas.metrics import Metric
from dbnd._core.tracking.schemas.tracking_info_run import RunInfo, ScheduledRunInfo
from dbnd._core.utils import json_utils
from dbnd._core.utils.dotdict import _as_dotted_dict
from dbnd._core.utils.timezone import utcnow
from dbnd._vendor._marshmallow.decorators import pre_load
from dbnd._vendor.marshmallow import fields, post_load
from dbnd._vendor.marshmallow_enum import EnumField
from dbnd.api.serialization.common import (
    ErrorInfoSchema,
    MetricSchema,
    TargetInfoSchema,
)
from dbnd.api.serialization.run import RunInfoSchema, ScheduledRunInfoSchema
from dbnd.api.serialization.task import TaskDefinitionInfoSchema, TaskRunInfoSchema
from dbnd.api.serialization.task_run_env import TaskRunEnvInfoSchema
from targets.data_schema import DataSchemaArgs, StructuredDataSchema


if typing.TYPE_CHECKING:
    from dbnd._core.tracking.schemas.tracking_info_objects import (
        ErrorInfo,
        TargetInfo,
        TaskDefinitionInfo,
        TaskRunEnvInfo,
        TaskRunInfo,
    )

AIRFLOW_SOURCE_TYPE = "airflow"


@attr.s
class TrackingSource(object):
    name = attr.ib()  # type: str
    url = attr.ib()  # type: str
    env = attr.ib()  # type: str

    source_type = attr.ib(default=AIRFLOW_SOURCE_TYPE)  # type: str
    source_instance_uid = attr.ib(default=None)  # type: Optional[UUID]


class TrackingSourceSchema(ApiStrictSchema):
    name = fields.String()
    url = fields.String()
    env = fields.String()

    source_type = fields.String()
    source_instance_uid = fields.UUID()

    @post_load
    def make_run_info(self, data, **kwargs):
        return TrackingSource(**data)


class AirflowTaskContextSchema(ApiStrictSchema):
    dag_id = fields.String()
    execution_date = fields.String()
    task_id = fields.String()
    try_number = fields.Integer(allow_none=True)
    airflow_instance_uid = fields.UUID()

    @post_load
    def make_run_info(self, data, **kwargs):
        return AirflowTaskContext(**data)


class TaskRunsInfoSchema(ApiStrictSchema):
    run_uid = fields.UUID()
    root_run_uid = fields.UUID()
    task_run_env_uid = fields.UUID()

    task_runs = fields.Nested(
        TaskRunInfoSchema, many=True, exclude=("task_signature_source",)
    )

    parent_child_map = fields.List(fields.List(fields.UUID()))
    upstreams_map = fields.List(fields.List(fields.UUID()))

    dynamic_task_run_update = fields.Boolean()

    targets = fields.Nested(TargetInfoSchema, many=True)
    task_definitions = fields.Nested(TaskDefinitionInfoSchema, many=True)
    af_context = fields.Nested(AirflowTaskContextSchema, allow_none=True)

    parent_task_run_uid = fields.UUID(allow_none=True)
    parent_task_run_attempt_uid = fields.UUID(allow_none=True)

    @post_load
    def make_run_info(self, data, **kwargs):
        return _as_dotted_dict(**data)


@attr.s
class InitRunArgs(object):
    root_run_uid = attr.ib()  # type: UUID
    run_uid = attr.ib()  # type: UUID
    task_runs_info = attr.ib()  # type: TaskRunsInfo
    driver_task_uid = attr.ib(default=None)  # type: UUID

    task_run_env = attr.ib(default=None)  # type: TaskRunEnvInfo

    new_run_info = attr.ib(
        default=None
    )  # type: Optional[RunInfo]  # we create it only for the new runs
    scheduled_run_info = attr.ib(default=None)  # type: Optional[ScheduledRunInfo]
    update_existing = attr.ib(default=False)  # type: bool
    source = attr.ib(default=UpdateSource.dbnd)  # type: UpdateSource
    af_with_monitor = attr.ib(default=True)  # type: bool
    af_context = attr.ib(default=None)  # type: Optional[AirflowTaskContext]
    tracking_source = attr.ib(default=None)  # type: Optional[TrackingSource]

    def asdict(self):
        return attr.asdict(self, recurse=False)

    @property
    def source_instance_uid(self) -> Optional[UUID]:
        """
        Currently (2022/01) this data is sent only when we have some known
        orchestration, i.e airflow or azkaban. If tracking data is sent from
        unknown/unsupported orchestrator (including also standalone script execution),
        or from dbnd orchestration - there will be no source_instance_uid.

        Will return source_instance_uid from any of specific contexts -
        either from airflow_context or from more generic tracking_source

        Will return None if no tracking_sources defined.
        """
        if self.af_context and self.af_context.airflow_instance_uid:
            return self.af_context.airflow_instance_uid

        if self.tracking_source and self.tracking_source.source_instance_uid:
            return self.tracking_source.source_instance_uid
        return None


class InitRunArgsSchema(ApiStrictSchema):
    run_uid = fields.UUID()
    root_run_uid = fields.UUID()

    driver_task_uid = fields.UUID(allow_none=True)

    task_run_env = fields.Nested(TaskRunEnvInfoSchema)
    task_runs_info = fields.Nested(TaskRunsInfoSchema)

    new_run_info = fields.Nested(RunInfoSchema, allow_none=True)
    scheduled_run_info = fields.Nested(ScheduledRunInfoSchema, allow_none=True)
    update_existing = fields.Boolean()
    af_with_monitor = fields.Boolean()
    source = fields.Str(allow_none=True)
    af_context = fields.Nested(AirflowTaskContextSchema, allow_none=True)
    tracking_source = fields.Nested(TrackingSourceSchema, allow_none=True)

    @post_load
    def make_init_run_args(self, data, **kwargs):
        return InitRunArgs(**data)


@attr.s
class TaskRunAttemptUpdateArgs(object):
    task_run_uid = attr.ib(repr=False)  # type: UUID
    task_run_attempt_uid = attr.ib(repr=False)  # type: UUID
    timestamp = attr.ib()  # type:  datetime
    state = attr.ib()  # type: TaskRunState
    error = attr.ib(default=None)  # type: Optional[ErrorInfo]
    attempt_number = attr.ib(default=None)  # type: int
    source = attr.ib(default=UpdateSource.dbnd)  # type: UpdateSource
    start_date = attr.ib(default=None)  # type: datetime
    external_links_dict = attr.ib(default=None)


class TaskRunAttemptUpdateArgsSchema(ApiStrictSchema):
    task_run_uid = fields.UUID()
    task_run_attempt_uid = fields.UUID()
    state = EnumField(TaskRunState, allow_none=True)
    timestamp = fields.DateTime(allow_none=True)
    error = fields.Nested(ErrorInfoSchema, allow_none=True)
    attempt_number = fields.Number(allow_none=True)
    source = fields.Str(allow_none=True)
    start_date = fields.DateTime(allow_none=True)
    external_links_dict = fields.Dict(allow_none=True)

    @post_load
    def make_object(self, data, **kwargs):
        return TaskRunAttemptUpdateArgs(**data)


class InitRunSchema(ApiStrictSchema):
    init_args = fields.Nested(InitRunArgsSchema)
    version = fields.String(allow_none=True)


init_run_schema = InitRunSchema()


class AddTaskRunsSchema(ApiStrictSchema):
    task_runs_info = fields.Nested(TaskRunsInfoSchema)
    source = EnumField(UpdateSource, allow_none=True)


add_task_runs_schema = AddTaskRunsSchema()


class SetDatabandRunStateSchema(ApiStrictSchema):
    run_uid = fields.UUID(required=True)
    state = EnumField(RunState)
    timestamp = fields.DateTime(required=False, default=None, missing=None)


set_run_state_schema = SetDatabandRunStateSchema()


class SetTaskRunReusedSchema(ApiStrictSchema):
    task_run_uid = fields.UUID(required=True)
    task_outputs_signature = fields.String(required=True)


set_task_run_reused_schema = SetTaskRunReusedSchema()


class UpdateTaskRunAttemptsSchema(ApiStrictSchema):
    task_run_attempt_updates = fields.Nested(TaskRunAttemptUpdateArgsSchema, many=True)


update_task_run_attempts_schema = UpdateTaskRunAttemptsSchema()


class SetUnfinishedTasksStateShcmea(ApiStrictSchema):
    run_uid = fields.UUID()
    state = EnumField(TaskRunState)
    timestamp = fields.DateTime()


set_unfinished_tasks_state_schema = SetUnfinishedTasksStateShcmea()


class SaveTaskRunLogSchema(ApiStrictSchema):
    task_run_attempt_uid = fields.UUID(required=True)
    log_body = fields.String(allow_none=True)
    local_log_path = fields.String(allow_none=True)


save_task_run_log_schema = SaveTaskRunLogSchema()


class SaveExternalLinksSchema(ApiStrictSchema):
    task_run_attempt_uid = fields.UUID(required=True)
    external_links_dict = fields.Dict(
        name=fields.Str(), url=fields.Str(), required=True
    )


save_external_links_schema = SaveExternalLinksSchema()


class LogMetricSchema(ApiStrictSchema):
    task_run_attempt_uid = fields.UUID(required=True)
    target_meta_uid = fields.UUID(allow_none=True)
    metric = fields.Nested(MetricSchema, allow_none=True)
    source = fields.String(allow_none=True)


log_metric_schema = LogMetricSchema()


class LogMetricsSchema(ApiStrictSchema):
    metrics_info = fields.Nested(LogMetricSchema, many=True)

    @post_load
    def make_object(self, data, **kwargs):
        data["metrics_info"] = [LogMetricArgs(**m) for m in data["metrics_info"]]


log_metrics_schema = LogMetricsSchema()


class DbtMetadataSchema(ApiStrictSchema):
    task_run_attempt_uid = fields.UUID(required=True)
    dbt_run_metadata = fields.Dict(required=True)


log_dbt_metadata_schema = DbtMetadataSchema()


class LogArtifactSchema(ApiStrictSchema):
    task_run_attempt_uid = fields.UUID(required=True)
    name = fields.String()
    path = fields.String()


log_artifact_schema = LogArtifactSchema()


@attr.s
class LogMetricArgs(object):
    task_run_attempt_uid = attr.ib()  # type: UUID
    metric = attr.ib()  # type: Metric
    source = attr.ib(default=None)  # type: Optional[str]
    target_meta_uid = attr.ib(default=None)  # type: Optional[UUID]

    def copy(self, key, value):
        return LogMetricArgs(
            task_run_attempt_uid=self.task_run_attempt_uid,
            source=self.source,
            target_meta_uid=self.target_meta_uid,
            metric=Metric(
                key=key,
                value=value,
                timestamp=self.metric.timestamp,
                source=self.metric.source,
            ),
        )


@attr.s
class LogDatasetArgs(object):
    run_uid = attr.ib()  # type: UUID
    task_run_uid = attr.ib()  # type: UUID
    task_run_name = attr.ib()  # type: str
    task_run_attempt_uid = attr.ib()  # type: UUID

    operation_path = attr.ib()  # type: Optional[str]

    operation_type = attr.ib()  # type: DbndDatasetOperationType
    operation_status = attr.ib()  # type: DbndTargetOperationStatus

    value_preview = attr.ib()  # type: Optional[str]
    data_dimensions = attr.ib()  # type: Optional[Tuple[Optional[int], Optional[int]]]
    data_schema = attr.ib()  # type: Optional[DataSchemaArgs]

    query = attr.ib(default=None)  # type: Optional[str]
    dataset_uri = attr.ib(default=None)  # type: Optional[str]
    columns_stats = attr.ib(default=attr.Factory(list))  # type: List[ColumnStatsArgs]

    operation_error = attr.ib(
        default=None
    )  # type: str # default=None :-> for compatibility with SDK < 51.0.0
    operation_source = attr.ib(default=None)  # type: Optional[str]
    timestamp = attr.ib(default=attr.Factory(utcnow))  # type: datetime
    with_partition = attr.ib(default=None)  # type: Optional[bool]

    def asdict(self):
        return attr.asdict(self, recurse=False)

    @property
    def records(self):
        return self.data_dimensions[0] if self.data_dimensions else None

    @property
    def columns(self):
        return self.data_dimensions[1] if self.data_dimensions else None

    @property
    def operation_status_value(self):
        return "SUCCESS" if self.operation_status.value == "OK" else "FAILED"


class LogDatasetSchema(ApiStrictSchema):
    run_uid = fields.UUID(required=True)
    task_run_uid = fields.UUID(required=True)
    task_run_name = fields.String(required=True)
    task_run_attempt_uid = fields.UUID(required=True)

    operation_path = fields.String(allow_none=True)
    dataset_uri = fields.String(allow_none=True, default=None)

    operation_type = EnumField(DbndDatasetOperationType)
    operation_status = EnumField(DbndTargetOperationStatus)
    operation_error = fields.String(allow_none=True)
    operation_source = fields.String(allow_none=True)
    timestamp = fields.DateTime(required=False, allow_none=True)

    value_preview = fields.String(allow_none=True)
    data_dimensions = fields.List(fields.Integer(allow_none=True), allow_none=True)
    data_schema = fields.Nested(StructuredDataSchema, allow_none=True)
    query = fields.String(allow_none=True)
    columns_stats = fields.Nested(ColumnStatsSchema, many=True, required=False)

    with_partition = fields.Boolean(required=False, allow_none=True)

    @pre_load
    def pre_load(self, data: dict) -> dict:
        data_schema = data.get("data_schema")
        if data_schema:
            # Backwared compatible with older SDK versions < v60.0.0
            if isinstance(data_schema, str):
                data["data_schema"] = json_utils.loads(data_schema)
        return data

    @post_load
    def make_object(self, data):
        return LogDatasetArgs(**data).asdict()


class LogDatasetsSchema(ApiStrictSchema):
    datasets_info = fields.Nested(LogDatasetSchema, many=True)


log_datasets_schema = LogDatasetsSchema()


@attr.s
class LogTargetArgs(object):
    run_uid = attr.ib()  # type: UUID
    task_run_uid = attr.ib()  # type: UUID
    task_run_name = attr.ib()  # type: str
    task_run_attempt_uid = attr.ib()  # type: UUID
    target_path = attr.ib()  # type: str
    operation_type = attr.ib()  # type: DbndTargetOperationType
    operation_status = attr.ib()  # type: DbndTargetOperationStatus

    value_preview = attr.ib()  # type: str
    data_dimensions = attr.ib()  # type: Optional[Tuple[Optional[int], Optional[int]]]
    data_schema = attr.ib()  # type: str
    data_hash = attr.ib()  # type: str

    param_name = attr.ib(default=None)  # type: Optional[str]
    task_def_uid = attr.ib(default=None)  # type: Optional[UUID]

    def asdict(self):
        return attr.asdict(self, recurse=False)


class LogTargetSchema(ApiStrictSchema):
    run_uid = fields.UUID(required=True)
    task_run_uid = fields.UUID(required=True)
    task_run_name = fields.String()
    task_run_attempt_uid = fields.UUID(required=True)

    target_path = fields.String()
    param_name = fields.String(allow_none=True)
    task_def_uid = fields.UUID(allow_none=True)
    operation_type = EnumField(DbndTargetOperationType)
    operation_status = EnumField(DbndTargetOperationStatus)

    value_preview = fields.String(allow_none=True)
    data_dimensions = fields.List(fields.Integer(), allow_none=True)
    data_schema = fields.String(allow_none=True)
    data_hash = fields.String(allow_none=True)


class LogTargetsSchema(ApiStrictSchema):
    targets_info = fields.Nested(LogTargetSchema, many=True)


log_targets_schema = LogTargetsSchema()


class HeartbeatSchema(ApiStrictSchema):
    run_uid = fields.UUID()


heartbeat_schema = HeartbeatSchema()


class AirflowTaskInfoSchema(ApiStrictSchema):
    execution_date = fields.DateTime()
    last_sync = fields.DateTime(allow_none=True)
    dag_id = fields.String()
    task_id = fields.String()
    task_run_attempt_uid = fields.UUID()
    retry_number = fields.Integer(required=False, allow_none=True)

    @post_load
    def make_object(self, data, **kwargs):
        return AirflowTaskInfo(**data)


class AirflowTaskInfosSchema(ApiStrictSchema):
    airflow_task_infos = fields.Nested(AirflowTaskInfoSchema, many=True)
    source = fields.String()
    base_url = fields.String()


airflow_task_infos_schema = AirflowTaskInfosSchema()


@attr.s
class TaskRunsInfo(object):
    run_uid = attr.ib()  # type: UUID
    root_run_uid = attr.ib()  # type: UUID

    task_run_env_uid = attr.ib()  # type: UUID

    task_runs = attr.ib(default=None)  # type: List[TaskRunInfo]
    task_definitions = attr.ib(default=None)  # type: List[TaskDefinitionInfo]
    targets = attr.ib(default=None)  # type: List[TargetInfo]

    parent_child_map = attr.ib(default=None)
    upstreams_map = attr.ib(default=None)
    dynamic_task_run_update = attr.ib(default=False)
    af_context = attr.ib(default=None)  # type: Optional[AirflowTaskContext]

    parent_task_run_uid = attr.ib(default=None)  # type: Optional[UUID]
    parent_task_run_attempt_uid = attr.ib(default=None)  # type: Optional[UUID]


@attr.s
class ScheduledJobInfo(object):
    uid = attr.ib()  # type: UUID
    name = attr.ib()  # type: str
    cmd = attr.ib()  # type: str
    start_date = attr.ib()  # type: datetime
    create_user = attr.ib()  # type: str
    create_time = attr.ib()  # type: datetime
    end_date = attr.ib(default=None)  # type: Optional[datetime]
    schedule_interval = attr.ib(default=None)  # type: Optional[str]
    catchup = attr.ib(default=None)  # type: Optional[bool]
    depends_on_past = attr.ib(default=None)  # type: Optional[bool]
    retries = attr.ib(default=None)  # type: Optional[int]
    active = attr.ib(default=None)  # type: Optional[bool]
    update_user = attr.ib(default=None)  # type: Optional[str]
    update_time = attr.ib(default=None)  # type: Optional[datetime]
    from_file = attr.ib(default=False)  # type: bool
    deleted_from_file = attr.ib(default=False)  # type: bool
    list_order = attr.ib(default=None)  # type: Optional[List[int]]
    job_name = attr.ib(default=None)  # type: Optional[str]


class ScheduledJobInfoSchema(ApiStrictSchema):
    uid = fields.UUID()
    name = fields.String()
    cmd = fields.String()
    start_date = fields.DateTime()
    create_user = fields.String()
    create_time = fields.DateTime()
    end_date = fields.DateTime(allow_none=True)
    schedule_interval = fields.String(allow_none=True)
    catchup = fields.Boolean(allow_none=True)
    depends_on_past = fields.Boolean(allow_none=True)
    retries = fields.Integer(allow_none=True)
    active = fields.Boolean(allow_none=True)
    update_user = fields.String(allow_none=True)
    update_time = fields.DateTime(allow_none=True)
    from_file = fields.Boolean(allow_none=True)
    deleted_from_file = fields.Boolean(allow_none=True)
    list_order = fields.Integer(allow_none=True)
    job_name = fields.String(allow_none=True)

    @post_load
    def make_object(self, data, **kwargs):
        return ScheduledJobInfo(**data)


class ScheduledJobArgsSchema(ApiStrictSchema):
    scheduled_job_args = fields.Nested(ScheduledJobInfoSchema)
    update_existing = fields.Boolean(default=False)


scheduled_job_args_schema = ScheduledJobArgsSchema()


@attr.s
class AirflowTaskInfo(object):
    execution_date = attr.ib()  # type: datetime
    dag_id = attr.ib()  # type: str
    task_id = attr.ib()  # type: str
    task_run_attempt_uid = attr.ib()  # type: UUID
    last_sync = attr.ib(default=None)  # type: Optional[datetime]
    retry_number = attr.ib(default=None)  # type: Optional[int]

    def asdict(self):
        return dict(
            execution_date=self.execution_date,
            last_sync=self.last_sync,
            dag_id=self.dag_id,
            task_id=self.task_id,
            task_run_attempt_uid=self.task_run_attempt_uid,
            retry_number=self.retry_number,
        )



def vcolor(data, pattern):
    """ Color a graph line by line

    :param data: the data
    :type data: list of tuples (info, value)
    :param pattern: list of colors, this list defines
      the pattern to color each line of the graph.
    :type pattern: list of 'colors' (str)
    :return: the colored graph
    :rtype: list of arrays (<info>, <value>, <color>)
    """
    ret = []
    l = len(pattern)
    c = 0
    for info, value in data:
        ret.append((info, value, pattern[c]))
        c = (c + 1) % l
    return ret


def hcolor(data, thresholds):
    """ Multicolor a graph according to thresholds

    :param data: the data
    :type data: list of tuples (info, value)
    :param thresholds: dict of thresholds, format
      {<threshold>: <color>,}
    :type thresholds: dict
    :return: the colored graph
    :rtype: list of arrays
    """
    ret = []
    for info, value in data:
        newval = []
        minover = None
        maxt = 0
        for t in thresholds:
            if maxt < t:
                maxt = t
            if value > t:
                newval.append((t, thresholds[t]))
            else:
                if minover is None or minover > t:
                    minover = t
        if minover is None:
            minover = maxt

        newval.append((value, thresholds[minover]))
        ret.append((info, newval))
    return ret



# © Copyright Databand.ai, an IBM Company 2022

from subprocess import CalledProcessError, call
from tempfile import NamedTemporaryFile


# based on https://stackoverflow.com/questions/13835055/python-subprocess-check-output-much-slower-then-call?lq=1
# standard check_output is superslow
def check_output(*popenargs, **kwargs):
    r"""Run command with arguments and return its output as a byte string.

    If the exit code was non-zero it raises a CalledProcessError.  The
    CalledProcessError object will have the return code in the returncode
    attribute and output in the output attribute.

    The arguments are the same as for the Popen constructor.  Example:
    """
    if "stdout" in kwargs:
        raise ValueError("stdout argument not allowed, it will be overridden.")

    with NamedTemporaryFile() as f:
        kwargs["stdout"] = f

        retcode = call(*popenargs, **kwargs)
        f.seek(0)
        output = f.read()
        if retcode:
            cmd = kwargs.get("args")
            if cmd is None:
                cmd = popenargs[0]
            raise CalledProcessError(retcode, cmd, output=output)
        return output



# © Copyright Databand.ai, an IBM Company 2022

import logging

from enum import Enum
from typing import Callable

import six

from dbnd._core.utils.basics.helpers import parse_bool
from dbnd._core.utils.basics.load_python_module import load_python_callable
from targets.types import NullableStr
from targets.values.value_type import ValueType


logger = logging.getLogger(__name__)

# custom types


# value types


class DataValueType(ValueType):
    """
    ValueType whose value is a ```target```.
    It will be "dereferenced accordingly to value_type
    """

    support_from_str = False
    load_on_build = False


class ObjectValueType(ValueType):
    type = object

    def is_handler_of_type(self, type_):
        # if user use future.str at py2, we still want to use this convertor
        if six.PY2 and type_.__name__ == "newobject":
            return True
        return super(ObjectValueType, self).is_handler_of_type(type_)


class DefaultObjectValueType(ObjectValueType):
    """
    used by func decorators to mark that it's "auto" type
    """


class StrValueType(ValueType):
    type = str

    type_str_extras = ("unicode",)
    support_merge = True

    def merge_values(self, *values):
        return "".join(values)

    def is_type_of(self, value):
        return isinstance(value, six.string_types)

    def is_handler_of_type(self, type_):
        # if user use future.str at py2, we still want to use this convertor
        if six.PY2 and type_.__name__ == "newstr":
            return True
        return super(StrValueType, self).is_handler_of_type(type_)

    def normalize(self, x):
        return str(x)


STR_VALUE_TYPE = StrValueType()


class NullableStrValueType(ValueType):
    """A ValueType that treats empty string as None"""

    type = NullableStr
    discoverable = False

    def to_str(self, x):
        if x is None:
            return ""
        else:
            return str(x)

    def parse_from_str(self, x):
        return x or None


class IntValueType(ValueType):
    type = int

    def parse_from_str(self, s):
        return int(s)

    def next_in_enumeration(self, value):
        return value + 1


class FloatValueType(ValueType):
    type = float

    def parse_from_str(self, s):
        """
        Parses a ``float`` from the string using ``float()``.
        """
        return float(s)


class EnumValueType(ValueType):
    type = Enum

    def __init__(self, enum):
        self._enum = enum
        super(EnumValueType, self).__init__()

    def parse_from_str(self, s):
        try:
            return self._enum[s]
        except KeyError:
            raise ValueError(
                "Invalid enum value '%s' - could not be parsed, use one of %s"
                % (s, list(self._enum))
            )

    def to_str(self, e):
        if e:
            return e.name
        return None


class BoolValueType(ValueType):
    type = bool

    def parse_from_str(self, s):
        return parse_bool(s)

    def normalize(self, value):
        return bool(value) if value is not None else None


class CallableValueType(ValueType):
    # should be used in configuration for now only
    type = Callable
    discoverable = False  # we can't really check if we got function

    def parse_from_str(self, s):
        return load_python_callable(callable_path=s)

    def to_str(self, x):
        return x.__name__



# © Copyright Databand.ai, an IBM Company 2022

import os

from sys import platform as _platform


is_osx = False
if _platform == "darwin":
    is_osx = True


def enable_osx_forked_request_calls():
    if not is_osx:
        return

    from dbnd._core.configuration.dbnd_config import config

    if not config.getboolean("core", "fix_env_on_osx"):
        return

    if "no_proxy" not in os.environ:
        os.environ["no_proxy"] = "*"

    if "OBJC_DISABLE_INITIALIZE_FORK_SAFETY" not in os.environ:
        os.environ["OBJC_DISABLE_INITIALIZE_FORK_SAFETY"] = "yes"



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.current import get_databand_context


def kill_run(run_uid, ctx=None):
    ctx = ctx or get_databand_context()
    return ctx.databand_api_client.api_request(endpoint="runs/stop", data=[run_uid])



# © Copyright Databand.ai, an IBM Company 2022

import datetime
import time

import six

from dbnd._vendor import pendulum
from dbnd._vendor.tabulate import tabulate


def patch_module_attr(module, name, value):
    original_name = "_original_" + name
    if getattr(module, original_name, None):
        return
    setattr(module, original_name, getattr(module, name))
    setattr(module, name, value)


def unpatch_module_attr(module, name, value):
    original_name = "_original_" + name
    if hasattr(module, original_name):
        setattr(module, name, getattr(module, original_name))


def patch_models(patches):
    for module, name, value in patches:
        patch_module_attr(module, name, value)


def unpatch_models(patches):
    for module, name, value in patches:
        unpatch_module_attr(module, name, value)


def tabulate_objects(
    objects,
    fields_filter=lambda f: not f.startswith("_"),
    headers=None,
    to_local_time=True,
):
    if not headers:
        headers = [k for k in dir(objects[0]) if fields_filter(k)]

    values = []
    for o in objects:
        o_values = []
        values.append(o_values)
        for h in headers:
            value = o.__getattribute__(h)
            if to_local_time and value and isinstance(value, datetime.datetime):
                value = pendulum.from_timestamp(
                    _timestamp(value), "local"
                ).to_rfc822_string()
            o_values.append(value)

    return tabulate(values, headers=headers)


def _timestamp(value):
    # type: (datetime.datetime)->float
    """backward support for timestamp in py2"""
    if six.PY3:
        return value.timestamp()

    return time.mktime(value.timetuple())


def safe_isinstance(obj, cls_str):
    """
    Checks whether a given obj implements the class who's name is in the cls_str parameter.
    This means that the isinstance check is performed without looking at the object's type hierarchy, but rather
    by class name.
    """
    if cls_str and obj:
        return cls_str in str(type(obj))
    return False



# © Copyright Databand.ai, an IBM Company 2022


class TargetError(Exception):
    """
    Base exception.
    """


class FileSystemException(TargetError):
    """
    Base class for generic file system exceptions.
    """


class FileAlreadyExists(FileSystemException):
    """
    Raised when a file system operation can't be performed because
    a directory exists but is required to not exist.
    """


class FailedToCopy(FileSystemException):
    """
    Raised when a copy operation failed
    """


class MissingParentDirectory(FileSystemException):
    """
    Raised when a parent directory doesn't exist.
    (Imagine mkdir without -p)
    """


class NotADirectory(FileSystemException):
    """
    Raised when a file system operation can't be performed because
    an expected directory is actually a file.
    """


class FileNotFoundException(FileSystemException):
    pass


class NotSupportedValue(TargetError):
    pass


class InvalidDeleteException(FileSystemException):
    pass
